/*	XGAppCore_mac.cpp
 *
 *		This is the application core class for the Macintosh.
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XError.h>
#include <XDataUtil.h>
#include <XCursorUtil.h>
#include <XApplication.h>
#include <XWindow.h>
#include <XMenu.h>
#include <XThread.h>

#include "factory.h"
#include "Copyright_mac.h"

#if OPT_FLOATWINDOW == 1
#include "MacFloater.h"
#endif

#include <Palettes.h>

/************************************************************************/
/*																		*/
/*	Global pointers														*/
/*																		*/
/************************************************************************/

XGAppCore					*_GApp = NULL;
EventRecord					_GCurEvent;

static bool					GHasAE;
static bool					GOS6Init;
static AEEventHandlerUPP	GAppleEventProc;

static XGIViewFactory		*GFactory;


/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGAppCore::XGAppCore
 *
 *		This attempts to initialize the Macintosh OS environment.
 *
 *		This simply initializes the MacOS environment, and installs the
 *	AEHandler routine, if we handle apple events
 */

XGAppCore::XGAppCore() : XGFocus(this)
{
	int i;
	long val;
	
	/*
	 *	Initialize view factory
	 */
	
	GFactory = new XGIViewFactory;
		
	/*
	 *	First, set the pointers up
	 */
	
	if (_GApp) throw XPostError("XGAppCore already created");
	_GApp = this;
	
	/*
	 *	Class initialization
	 */
	
	fQuitFlag = false;
	_fModalWindow = NULL;
	fFront = true;
	
	/*
	 *	Bring up the Macintosh OS
	 */
	
	for (i = 0; i < 10; i++) MoreMasters();		// Grow the masters first
	
	InitGraf(&qd.thePort);						// initialize the grafport
	InitFonts();								// Init the font manager
	FlushEvents(everyEvent,0);					// Flush the events
	InitWindows();								// init window manager
	InitMenus();								// initialize menu manager
	TEInit();									// text edit initialize
	InitDialogs(NULL);							// Init dialog manager
	InitCursor();								// Init cursor engine
		
	/*
	 *	Initialize apple events (if we have 'em)
	 */
	
	GHasAE = false;
	if (Gestalt(gestaltAppleEventsAttr,&val)) val = 0;
	else if (val & 1) GHasAE = true;
	
	if (GHasAE) {
		// We have apple events. Initialize
		GAppleEventProc = NewAEEventHandlerProc(AEHandleProc);
		
		AEInstallEventHandler(kCoreEventClass,kAEOpenApplication,GAppleEventProc,0,false);
		AEInstallEventHandler(kCoreEventClass,kAEOpenDocuments,GAppleEventProc,0,false);
		AEInstallEventHandler(kCoreEventClass,kAEPrintDocuments,GAppleEventProc,0,false);
		AEInstallEventHandler(kCoreEventClass,kAEQuitApplication,GAppleEventProc,0,false);
	} else {
		// If this is 'false', we call InitOS6Finder as soon as Run is called
		GOS6Init = false;
	}
}

/*	XGAppCore::~XGAppCore
 *
 *		Shut down. This kills the application core stuff
 */

XGAppCore::~XGAppCore()
{
	if (_GApp != this) return;
	
	/*
	 *	Now, shut down and release allocated stuff
	 */
	
	if (GHasAE) {
		DisposeRoutineDescriptor(GAppleEventProc);
	}
	if (GFactory) delete GFactory;
	
	/*
	 *	Last, remove me from pointer
	 */
	
	_GApp = NULL;
}

/************************************************************************/
/*																		*/
/*	MacOS callbacks														*/
/*																		*/
/************************************************************************/

/*	XGAppCore::Open
 *
 *		Default: do nothing
 */

long XGAppCore::Open(FSSpec *)
{
	return 0;
}

/*	XGAppCore::Print
 *
 *		Default: do nothing
 */

long XGAppCore::Print(FSSpec *)
{
	return 0;
}

/*	XGAppCore::New
 *
 *		Default: do nothing
 */

long XGAppCore::New(void)
{
	return 0;
}

/*	XGAppCore::Quit
 *
 *		Default: call the Exit() routine for an orderly shutdown
 */

long XGAppCore::Quit(void)
{
	WindowPtr w;
	EventRecord e;
	XGWindow *tw;
	short dirty;
		
	/*
	 *	Run a mini-event loop, updating and activating things only.
	 *	The purpose of this is to allow the user to see what he is closing
	 *	when he is prompted...
	 */
	
	for (;;) {

#if OPT_FLOATWINDOW == 1
		w = MFrontWindow();
#else
		w = FrontWindow();
#endif

		for (;;) {
			if (w == NULL) break;				// Success.
			tw = XGWindow::_GetXGWindow(w);
			if ((tw == NULL) || (!(tw->_Attributes() & KAttrNoClose))) break;
			w = (WindowPtr)(((WindowPeek)w)->nextWindow);
		}
		if (w == NULL) break;					// Run to next
		if (tw == NULL) {
			if (!DoClose(w)) return 0;
			continue;
		}
		dirty = tw->IsDirty();
		if (!dirty) {
			if (!DoClose(w)) return 0;		// Unable to close window...
			continue;						// Quick short circuit mechanism
		}
		
		/*
		 *	Process events until there is a NULL event (which means
		 *	things are stable again). This allows the windows a chance
		 *	to be updated in such a way that the user can see what's going
		 *	on inside of them.
		 */
		
		for (;;) {
			GetNextEvent(everyEvent,&e);
			if (e.what == nullEvent) break;
			if (e.what == updateEvt) DoUpdate(&e);
		}
		
		/*
		 *	See if the window can be closed now...
		 */
		
		if (!DoClose(w)) return 0;			// Unable to close window
	}
	
	/*
	 *  Once we've been here, this closes the unclosable windows
	 */
	
	for (;;) {
#if OPT_FLOATWINDOW == 1
		w = MFrontWindow();
#else
		w = FrontWindow();
#endif
		if (w == NULL) break;
		tw = XGWindow::_GetXGWindow(w);
		
		if (tw == NULL) {
			DoClose(w);
		} else {
			delete tw;
		}
	}
	
	/*
	 *  Now, last but not least, close the palette windows
	 */
	
	for (;;) {
		w = FrontWindow();
		if (w == NULL) break;
		tw = XGWindow::_GetXGWindow(w);
		
		if (tw == NULL) {
			DoClose(w);
		} else {
			delete tw;
		}
	}
		
	/*
	 *  Success. Terminate.
	 */
	 
	Exit();
	return 1;									// Success
}

/*	XGAppCore::FilterEvent
 *
 *		This gives the application the opportunity to filter out a
 *	message. If this returns 'true', the message is skipped
 */

bool XGAppCore::FilterEvent(EventRecord *)
{
	return false;
}

/************************************************************************/
/*																		*/
/*	Event Loop head														*/
/*																		*/
/************************************************************************/

/*	XGAppCore::_RunOnce
 *
 *		This is an internal routine used for modal window support; this
 *	runs an event loop once with the specified window as the modal window
 */

bool XGAppCore::_RunOnce(XGWindow *w)
{
	RgnHandle rgn;
	
	_fModalWindow = w;					// I'm the modal window
	
	/*
	 *	Do event loop once
	 */
	
	rgn = NewRgn();
	EmptyRgn(rgn);
	
	XGThread::YieldThread();
	XGBroadcast::_SendPost();
	XGBroadcast::SendBroadcast(KEventPrePeriodic,0,NULL);
	WaitNextEvent(everyEvent,&_GCurEvent,GetDblTime(),rgn);
	_GetCurFocus()->SendDispatch(KEventIdle,0,NULL);
	if (!FilterEvent(&_GCurEvent)) {
		DoEvent(&_GCurEvent);
		if (!fQuitFlag) HiliteMenu(0);
	}
	XGBroadcast::SendBroadcast(KEventPostPeriodic,0,NULL);
	
	DisposeRgn(rgn);
	
	/*
	 *	Done. Clean up and return
	 */

	_fModalWindow = NULL;				// No longer modal window
	return fQuitFlag;
}

/*	XGAppCore::Run
 *
 *		Run the event loop.
 */

void XGAppCore::Run(void)
{
	RgnHandle rgn;
	
	/*
	 *	When we start up, parse the finder information (for System v6)
	 */
	
	if (!GOS6Init) {
		InitOS6Finder();
		GOS6Init = true;
	}
	
	/*
	 *	Initialize some temporary stuff
	 */
	
	rgn = NewRgn();
	EmptyRgn(rgn);
	
	/*
	 *	Now run my event loop
	 */
	
	XGBroadcast::SendBroadcast(KEventPostPeriodic,0,NULL);
	while (!fQuitFlag) {
		// Thread support
		XGThread::YieldThread();
		XGBroadcast::_SendPost();
			
		// Send the idle event
		XGBroadcast::SendBroadcast(KEventPrePeriodic,0,NULL);

		// Get the macintosh event
		WaitNextEvent(everyEvent,&_GCurEvent,GetDblTime(),rgn);

		// Get the idle event loop
		_GetCurFocus()->SendDispatch(KEventIdle,0,NULL);
		
		// Filter this event
		if (!FilterEvent(&_GCurEvent)) {
			_fModalWindow = NULL;			// make sure this is clear
			DoEvent(&_GCurEvent);
			if (!fQuitFlag) HiliteMenu(0);
		}
		
		// Send the post idle event
		XGBroadcast::SendBroadcast(KEventPostPeriodic,0,NULL);
	}
	
	/*
	 *	Release temporary memory
	 */
	
	DisposeRgn(rgn);
}

/*	XGAppCore::Exit
 *
 *		This forces the event loop to exit
 */

void XGAppCore::Exit()
{
	fQuitFlag = true;
}

/************************************************************************/
/*																		*/
/*	Root Event Processing												*/
/*																		*/
/************************************************************************/

/*	XGAppCore::ReceiveDispatch
 *
 *		Receive this dispatched message; this handles menu commands and
 *	the like
 */

long XGAppCore::ReceiveDispatch(long msg, long arg, void *parg)
{
	WindowPtr w;
	short i;
	XGSMenuStatusRecord *gm;
	
	switch (msg) {
		case KEventDoMenuCommand:
			/*
			 *	Handle the following menu commands:
			 *
			 *		Quit/Close/About
			 *		Undo/Cut/Copy/Paste/Clear (desk accessories)
			 */
			
			if (arg == KMenuFileQuit) {
				if (!IsModalWindow()) {
					Quit();
					return 1;
				}
			}
			if (arg == KMenuAboutMe) {
				ShowCopyright();
				return 1;
			}
			
#if OPT_FLOATWINDOW == 1
			w = MFrontWindow();
#else
			w = FrontWindow();
#endif

			if (w == NULL) return 0;
			if (arg == KMenuFileClose) {
				if (!IsModalWindow()) {
				
#if OPT_FLOATWINDOW == 1
					DoClose(MFrontWindow());
#else
					DoClose(FrontWindow());
#endif

					return 1;
				}
			}
			
			if (IsModalWindow()) break;		// skip the rest
			
			i = ((WindowPeek)w)->windowKind;
			if (i == dialogKind) {
				switch (arg) {
					case KMenuEditCut:
						DialogCut(w);
						return 1;
					case KMenuEditCopy:
						DialogCopy(w);
						return 1;
					case KMenuEditPaste:
						DialogPaste(w);
						return 1;
					case KMenuEditClear:
						DialogDelete(w);
						return 1;
				}
			} else if (i < 0) {
				switch (arg) {
					case KMenuEditUndo:
						SystemEdit(0);
						return 1;
					case KMenuEditCut:
						SystemEdit(2);
						return 1;
					case KMenuEditCopy:
						SystemEdit(3);
						return 1;
					case KMenuEditPaste:
						SystemEdit(4);
						return 1;
					case KMenuEditClear:
						SystemEdit(5);
						return 1;
				}
			}
			break;
		
		case KEventGetMenuStatus:
			/*
			 *	Handle enabling those I can handle
			 */
			
			gm = (XGSMenuStatusRecord *)parg;
			
			if (!IsModalWindow() && (arg == KMenuFileQuit)) {
				gm->enable = true;
				return 1;
			}
			if (arg == KMenuAboutMe) {
				gm->enable = true;
				return 1;
			}
			
			if (IsModalWindow()) break;			// skip the rest
			
#if OPT_FLOATWINDOW == 1
			w = MFrontWindow();
#else
			w = FrontWindow();
#endif

			if (w == NULL) return 0;
			if (arg == KMenuFileClose) {
				gm->enable = true;
				return 1;
			}
			
			i = ((WindowPeek)w)->windowKind;
			
			if (i == dialogKind) {
				switch (arg) {
					case KMenuEditCut:
					case KMenuEditCopy:
					case KMenuEditPaste:
					case KMenuEditClear:
						gm->enable = true;
						return 1;
				}
			} else if (i < 0) {
				switch (arg) {
					case KMenuEditUndo:
					case KMenuEditCut:
					case KMenuEditCopy:
					case KMenuEditPaste:
					case KMenuEditClear:
						gm->enable = true;
						return 1;
				}
			}
			break;
	
		default:
			break;
	}
	
	return XGDispatch::ReceiveDispatch(msg,arg,parg);
}


/************************************************************************/
/*																		*/
/*	Corner Icon Support													*/
/*																		*/
/************************************************************************/

/*	XGAppCore::InvalidateCornerIcon
 *
 *		This invalidates the corner icon object
 */

void XGAppCore::InvalidateCornerIcon(WindowPtr w)
{
	Rect r;
	
	SetPort(w);
	r = w->portRect;
	r.left = r.right - 15;
	r.top = r.bottom - 15;
	SetOrigin(0,0);
	InvalRect(&r);
}

/*	XGAppCore::DrawCornerIcon
 *
 *		This draws the grow icon, clipping the area appropriately
 */
 
void XGAppCore::DrawCornerIcon(WindowPtr w)
{
	Rect r;
	
	SetPort(w);
	SetOrigin(0,0);
	r = w->portRect;
	r.left = r.right - 15;
	r.top = r.bottom - 15;
	ClipRect(&r);
	DrawGrowIcon(w);
	ClipRect(&(w->portRect));
}

/************************************************************************/
/*																		*/
/*	Drag/Drop Support													*/
/*																		*/
/************************************************************************/

/*	XGAppCore::InitOS6Finder
 *
 *		This is the code to initialize the environment for the MacOS v6
 *	(pre-AppleEvent) Finder
 */

void XGAppCore::InitOS6Finder()
{
#ifndef powerc					// Not on the PowerPC
	short what,num;
	AppFile f;
	short i;
	FSSpec file;
	
	CountAppFiles(&what,&num);
	if (num == 0) {
		/*
		 *	New application.
		 */
		
		_GApp->New();
	} else if (what == 0) {
		/*
		 *	Open specified files.
		 */
		
		for (i = 1; i <= num; i++) {
			GetAppFiles(i,&f);
			file.vRefNum = f.vRefNum;
			file.parID = 0;
			pstrcpy(file.name,f.fName);
			_GApp->Open(&file);
		}
	} else {
		/*
		 *	Print specified files and quit
		 */
		
		for (i = 1; i <= num; i++) {
			GetAppFiles(i,&f);
			file.vRefNum = f.vRefNum;
			file.parID = 0;
			pstrcpy(file.name,f.fName);
			_GApp->Print(&file);
		}
		_GApp->Quit();
	}
#endif // powerc
}

/*	XGAppCore::GotRequiredParams
 *
 *		Check to see all the required parameters were fetched
 */

OSErr XGAppCore::GotRequiredParams(AppleEvent *event)
{
	OSErr myErr;
	long size;
	DescType returnType;
	
	myErr = AEGetAttributePtr(event,keyMissedKeywordAttr,typeWildCard,&returnType,NULL,0,&size);
	if (myErr == noErr) return errAEParamMissed;
	else if (myErr == errAEDescNotFound) return noErr;
	else return myErr;
}

/*	XGAppCore::AEHandleProc
 *
 *		This is the code to handle the Apple Events for MacOS7 and later
 */

pascal OSErr XGAppCore::AEHandleProc(AppleEvent *event, AppleEvent */*ret*/, long)
{
	OSErr myErr;
	long eClass,eType;
	DescType returnType;
	long size;
	long off,len;
	AEDescList docList;
	AEKeyword keyWord;
	FSSpec file;
	
	/*
	 *	Get the Apple Event message type
	 */
	
	if (noErr != (myErr = AEGetAttributePtr(event,keyEventClassAttr,typeType,&returnType,(Ptr)&eClass,sizeof(eClass),&size))) return myErr;
	if (noErr != (myErr = AEGetAttributePtr(event,keyEventIDAttr,typeType,&returnType,(Ptr)&eType,sizeof(eType),&size))) return myErr;
	
	/*
	 *	We've now got the class and type of the event. Dispatch for the
	 *	common four entry points.
	 */
	
	if (eClass == kCoreEventClass) {
		switch (eType) {
			case kAEOpenApplication:
				/*
				 *	New application event
				 */
				
				myErr = GotRequiredParams(event);
				if (myErr == noErr) _GApp->New();
				return myErr;
			case kAEOpenDocuments:
				/*
				 *	Open new documents
				 */
				
				if (noErr != (myErr = AEGetParamDesc(event,keyDirectObject,typeAEList,&docList))) return myErr;
				if (noErr != (myErr = AECountItems(&docList,&len))) {
					AEDisposeDesc(&docList);
					return myErr;
				}
				for (off = 1; off <= len; off++) {
					if (noErr != (myErr = AEGetNthPtr(&docList,off,typeFSS,&keyWord,&returnType,(Ptr)&file,sizeof(file),&size))) {
						AEDisposeDesc(&docList);
						return myErr;
					}
					
					_GApp->Open(&file);
				}
				AEDisposeDesc(&docList);
				break;
			case kAEPrintDocuments:
				/*
				 *	Print specified documents
				 */
				
				if (noErr != (myErr = AEGetParamDesc(event,keyDirectObject,typeAEList,&docList))) return myErr;
				if (noErr != (myErr = AECountItems(&docList,&len))) {
					AEDisposeDesc(&docList);
					return myErr;
				}
				for (off = 1; off <= len; off++) {
					if (noErr != (myErr = AEGetNthPtr(&docList,off,typeFSS,&keyWord,&returnType,(Ptr)&file,sizeof(file),&size))) {
						AEDisposeDesc(&docList);
						return myErr;
					}
					
					_GApp->Print(&file);
				}
				AEDisposeDesc(&docList);
				break;
			case kAEQuitApplication:
				/*
				 *	Quit this application
				 */
				
				myErr = GotRequiredParams(event);
				if (myErr == noErr) return (_GApp->Quit()) ? noErr : 1;
				else return myErr;
			default:
				return errAEEventNotHandled;			// Not able to handle event
		}
	
		return GotRequiredParams(event);				// Fall through to handle this		
	}
	
	/*
	 *	Process the rest of the possible events
	 */
	
	return errAEEventNotHandled;
}


/************************************************************************/
/*																		*/
/*	Event Loop Dispatch													*/
/*																		*/
/************************************************************************/

/*	XGAppCore::DoEvent
 *
 *		This is the top of the traditional Macintosh event loop. This
 *	dispatches to the correct routine depending on what's
 */

void XGAppCore::DoEvent(EventRecord *e)
{
	OSErr err;
	
	/*
	 *	Dispatch events
	 */
	
	try {
		DoMouseMove(e);
		switch (e->what) {
			case nullEvent:
				break;
			case updateEvt:
				DoUpdate(e);
				break;
			case activateEvt:	// If OPT_FLOATWINDOW is true, activation elsewhere
#if OPT_FLOATWINDOW == 0
				_DoActive((WindowPtr)(e->message),(e->modifiers) & 1);
#endif
				break;
			case mouseDown:
				DoMouse(e);
				break;
			case keyDown:
			case autoKey:
				if (e->modifiers & cmdKey) {
					if (e->what == keyDown) {
						DoMenu(MenuKey((char)(e->message & charCodeMask)));
					}
				} else {
					DoKeyDown(e);
				}
				break;
			
			case diskEvt:
				/* Handle disk insertion event right here */
				if (GETHIWORD(e->message) != 0) {
					Point pt;
					
					DILoad();
					pt.h = 120;
					pt.v = 120;
					DIBadMount(pt,e->message);
					DIUnload();
				}
				break;
				
			case osEvt:
				DoOSEvent(e);
				break;
			case kHighLevelEvent:
				err = AEProcessAppleEvent(e);
				if (err) {
					NormalCursor();
					throw XPostError("Apple Event error occured");
				}
				break;
		}
	}
	
	catch (XError &e) {
		e.DisplayError();
	}
}

/************************************************************************/
/*																		*/
/*	Mouse Event Processing												*/
/*																		*/
/************************************************************************/

/*	ExtractKeyState
 *
 *		Given a key array and a key code, this returns 1 if the specified
 *	key is down
 */

static int ExtractKeyState(unsigned char *keys, short code)
{
	return ((keys[code >> 3] >> (code & 0x07)) & 1);
}

/*	XGAppCore::DoMouseMove
 *
 *		Handle a moving mouse moving across one of my windows
 */

void XGAppCore::DoMouseMove(EventRecord *e)
{
	int i;
	WindowPtr w;
	XGWindow *ww;
	Point pt;
		
	i = FindWindow(e->where,&w);
	if (i != inContent) {
		NormalCursor();
		
		/*
		 *  Do the balloon help for the resize thing
		 */
		
#ifdef OLD_BALLOON
		if (i == inGrow) {
			BalloonRecord record;
			Point pt;
			
			record.majorID = 'NGUI';
			record.minorID = 'SIZE';
			record.state = -1;
			record.args = NULL;
			SetPort(w);
			SetOrigin(0,0);
			record.area = w->portRect;
			record.area.top = record.area.bottom - 15;
			record.area.left = record.area.right - 15;
			pt.h = 0;
			pt.v = 0;
			LocalToGlobal(&pt);
			OffsetRect(&record.area,pt.h,pt.v);
			TView::DisplayHelpBalloon(&record);		// Display it here...
		} else {
			TView::HideHelpBalloon();				// No help for me now.
		}
#endif

	} else {
		ww = XGWindow::_GetXGWindow(w);
		
		// modal window screen
		if (_fModalWindow) {
			if ((ww != _fModalWindow) && (ww != NULL)) return;
		}
		
		if (ww == NULL) {
			NormalCursor();
		} else {
			unsigned char keys[16];
			
			/*
			 *	Macintosh OS workaround--
			 *
			 *		The mouse move event from the osEvt event
			 *	does not put the modifiers field correctly. This extracts
			 *	the actual key bindings and dispatches with those instead.
			 */
			
			GetKeys((UInt32 *)keys);
			i = 0;
			
			if (ExtractKeyState(keys,0x38) || ExtractKeyState(keys,0x3C)) i |= KKeyShift;
			if (ExtractKeyState(keys,0x37)) i |= KKeyControl;
			if (ExtractKeyState(keys,0x3B) || ExtractKeyState(keys,0x3E)) i |= KKeyCommand;
			if (ExtractKeyState(keys,0x3A) || ExtractKeyState(keys,0x3D)) i |= KKeyOption;
			
			ww->_ResetDrawing();
			pt = e->where;
			GlobalToLocal(&pt);
			ww->_ProcessMouseMove(pt,i);
		}
	}
}

/*	XGAppCore::DoMouseDown
 *
 *		Handle mouse down event inside of the specified window. This tracks
 *	the mouse move stuff as well.
 */
 
void XGAppCore::DoMouseDown(WindowPtr w, Point where, short modifiers)
{
	XGWindow *ww;
	short flags;
	short x,y;
	long time;
	Point last;
	
	/*
	 *	Reset and do the coordinate transform thing
	 */
	
	ww = XGWindow::_GetXGWindow(w);
	if (ww == NULL) return;
	
	ww->_ResetDrawing();
	SetPort(w);
	GlobalToLocal(&where);
	
	/*
	 *	Mouse was down in the content area. Process it as such
	 */
	
	flags = KKeyButton;
	if (modifiers & shiftKey) flags |= KKeyShift;
	if (modifiers & controlKey) flags |= KKeyCommand;
	if (modifiers & cmdKey) flags |= KKeyControl;
	if (modifiers & optionKey) flags |= KKeyOption;

	/*
	 *	Double click detection
	 */
	
	x = where.h - fLastClickPos.h;
	y = where.v - fLastClickPos.v;
	if ((-4 < x) && (x < 4) && (-4 < y) && (y < 4)) {
		time = TickCount();
		if (time - fLastClickTime <= GetDblTime()) {
			flags |= KDoubleClick;
		}
	}
	
	/*
	 *	Mouse down stuff
	 */
	 
	if (ww->_ProcessMouseDown(where,flags)) {
		/*
		 *	User captured mouse. Do the right things
		 */
		
		XGBroadcast::SendBroadcast(KEventPostPeriodic,0,NULL);
		while (Button()) {
			ww->_ResetDrawing();
			GetMouse(&last);
			if ((last.h != where.h) || (last.v != where.v)) {
				XGBroadcast::SendBroadcast(KEventPrePeriodic,0,NULL);
				_GetCurFocus()->SendDispatch(KEventIdle,0,NULL);
				ww->_ProcessMouseMove(last,flags | KCapturedMouse);
				XGBroadcast::SendBroadcast(KEventPostPeriodic,0,NULL);
				where = last;
			}
		}
		XGBroadcast::SendBroadcast(KEventPrePeriodic,0,NULL);
		_GetCurFocus()->SendDispatch(KEventIdle,0,NULL);
		ww->_ProcessMouseUp(where,flags | KCapturedMouse);
	}
	
	/*
	 *	Store information for mouse double click detection
	 */
	
	if (flags & KDoubleClick) {
		fLastClickTime -= GetDblTime();		// assure out of time
	} else {
		fLastClickTime = TickCount();
		fLastClickPos = where;
	}
}

/*	XGAppCore::DoMouse
 *
 *		Handle mouse down event
 */

void XGAppCore::DoMouse(EventRecord *e)
{
	short i;
	WindowPtr w;
	XGWindow *ww;
	Rect r;
	GrafPtr foo;
	long l;
	Point min,max;
	Point zoom;
	
	i = FindWindow(e->where,&w);				// Get current window
	ww = XGWindow::_GetXGWindow(w);				// Get XGWindow object
	
	// screen for modal window
	if (_fModalWindow) {
		if ((ww != _fModalWindow) && (ww != NULL)) {
			SysBeep(5);
			return;
		}
	}
	
	// dispatch message
	switch (i) {
		case inDesk:							// In the desk top
			break;
		case inMenuBar:							// In the menu bar
			DoMenu(MenuSelect(e->where));
			break;
		case inSysWindow:						// In a desk accessory
			SystemClick(e,w);
			break;
		case inContent:							// Click in content
		
#if OPT_FLOATWINDOW == 1
			if (w == MFrontWindow()) DoMouseDown(w,e->where,e->modifiers);
			else if (w == FrontWindow()) DoMouseDown(w,e->where,e->modifiers);
			else MSelectWindow(w);
#else
			if (w == FrontWindow()) DoMouseDown(w,e->where,e->modifiers);
			else SelectWindow(w);
#endif

			break;
		case inGoAway:
			if (TrackGoAway(w,e->where)) DoClose(w);
			break;
		case inDrag:
			GetWMgrPort(&foo);
			r = foo->portRect;
			InsetRect(&r,2,2);
			r.top += LMGetMBarHeight();

#if OPT_FLOATWINDOW == 1
			MDragWindow(w,e->where,&r);
#else
			DragWindow(w,e->where,&r);
#endif
			break;
		case inGrow:
			if (ww != NULL) {
				min = ww->MinWindowSize();
				ww->_SizeToContent(&min);
				if (min.h == -1) min.h = 125;
				if (min.v == -1) min.v = 75;
				
				max = ww->MaxWindowSize();
				ww->_SizeToContent(&max);
				if (max.h == -1) max.h = 32000;
				if (max.v == -1) max.v = 32000;
				
				SetRect(&r,min.h,min.v,max.h+1,max.v+1);
			} else {
				SetRect(&r,125,75,32000,32000);
			}
			if (0 != (l = GrowWindow(w,e->where,&r))) {
				if (ww) {
					ww->_ResetDrawing();
					InvalidateCornerIcon(w);		// Invalidate in old loc
				} else {
					SetPort(w);
					ClipRect(&(w->portRect));
					EraseRect(&(w->portRect));
				}
				SizeWindow(w,(short)(l&0x0FFFF),(short)(l>>16),1);
				if (ww) {
					ww->_ResetDrawing();
					InvalidateCornerIcon(w);		// Invalidate in new loc
					ww->_ProcessSizeEvent();
				} else {
					ClipRect(&(w->portRect));
					InvalRect(&(w->portRect));
				}
			}
			break;
		case inZoomIn:
		case inZoomOut:
			if (TrackBox(w,e->where,i)) {
//###				Rect s;
				
				if (ww != NULL) {
					/*  Get zoom size right for this window  */
					zoom = ww->ZoomWindowSize();
					
					ww->_SizeToContent(&zoom);
					r = ((WStateData *)*(((WindowPeek)w)->dataHandle))->stdState;
					if (zoom.h == -1) zoom.h = r.right - r.left;
					if (zoom.v == -1) zoom.v = r.bottom - r.top;
					ww->_ContentToSize(&zoom);
					ww->_LimitWindowSize(&zoom);
					ww->_SizeToContent(&zoom);
					
					if (zoom.h != -1) {
						if (r.right - r.left > zoom.h) r.right = r.left + zoom.h;
					}
					if (zoom.v != -1) {
						if (r.bottom - r.top > zoom.v) r.bottom = r.top + zoom.v;
					}
					
					/* Bound inside legal area (if there is one) */
//###					_GApp->WindowBoundary(&s);
//					if (!EmptyRect(&s)) {
//						s.top += 21;
//						s.bottom -= 1;
//						s.left += 1;
//						s.right -= 1;
//						
//						if (r.left < s.left) {
//							r.right += s.left - r.left;
//							if (r.right > s.right) r.right = s.right;
//							r.left = s.left;
//						}
//						if (r.right > s.right) {
//							r.left += s.right - r.right;
//							if (r.left < s.left) r.left = s.left;
//							r.right = s.right;
//						}
//						if (r.top < s.top) {
//							r.bottom += s.top - r.top;
//							if (r.bottom > s.bottom) r.bottom = s.bottom;
//							r.top = s.top;
//						}
//						if (r.bottom > s.bottom) {
//							r.top += s.bottom - r.bottom;
//							if (r.top < s.top) r.top = s.top;
//							r.bottom = s.bottom;
//						}
//					}
					
					/* Update the zoom state */
					((WStateData *)*(((WindowPeek)w)->dataHandle))->stdState = r;
					
					SetPort(w);
					SetOrigin(0,0);
					ClipRect(&(w->portRect));
					EraseRect(&(w->portRect));
					ZoomWindow(w,i,0);
					ClipRect(&(w->portRect));
					InvalRect(&(w->portRect));
					ww->_ProcessSizeEvent();
				} else {
//###					GApplication->WindowBoundary(&s);
//					if (!EmptyRect(&s)) {
//						s.top += 21;
//						s.bottom -= 1;
//						s.left += 1;
//						s.right -= 1;
//						((WStateData *)*(((WindowPeek)w)->dataHandle))->stdState = s;
//					}
					
					SetPort(w);
					ClipRect(&(w->portRect));
					EraseRect(&(w->portRect));
					ZoomWindow(w,i,0);
					ClipRect(&(w->portRect));
					InvalRect(&(w->portRect));
				}
			}
			break;
	}
}

/************************************************************************/
/*																		*/
/*	Miscellaneous Stuff													*/
/*																		*/
/************************************************************************/

/*	XGAppCore::DoUpdate
 *
 *		Do the update event
 */

void XGAppCore::DoUpdate(EventRecord *e)
{
	XGWindow *ww;
	WindowPtr w = (WindowPtr)(e->message);
	RgnHandle rgn;
	
	ww = XGWindow::_GetXGWindow(w);
	if (ww) ww->_ResetDrawing();
	SetPort(w);
	ClipRect(&(w->portRect));
	
	BeginUpdate(w);
	if (ww) {
		if (!(ww->_Attributes() & KAttrNoResize)) DrawCornerIcon(w);
		rgn = w->visRgn;
		ww->_ProcessUpdateEvent((**rgn).rgnBBox);
		if (!(ww->_Attributes() & KAttrNoResize)) DrawCornerIcon(w);	// Make sure it's okay
	}
	EndUpdate(w);
}

#if OPT_FLOATWINDOW == 1

/*	DoActivate
 *
 *		Trap from MacFloater.c to me
 */

void DoActivate(WindowPtr w, short flag)
{
	_GApp->_DoActive(w,flag);
}

#endif

/*	XGAppCore::_DoActive
 *
 *		handle the activation event
 */

void XGAppCore::_DoActive(WindowPtr w, short flag)
{
	XGWindow *ww;
	long version;
	
	ww = XGWindow::_GetXGWindow(w);
	::Gestalt(gestaltQuickdrawVersion,&version);
	if (version >= 0x100) ::ActivatePalette(w);
	if (ww) {
		if (!(ww->_Attributes() & KAttrNoResize)) DrawCornerIcon(w);
		ww->_ResetDrawing();
		
		ww->_ProcessActivateEvent(flag);
		if (!(ww->_Attributes() & KAttrNoResize)) DrawCornerIcon(w);
	}
}

/*	XGAppCore::DoKeyDown
 *
 *		Handle keyboard event
 */

void XGAppCore::DoKeyDown(EventRecord *e)
{
	unsigned char key;
	unsigned char code;
	short mods;
	XGFocus *f = _GetCurFocus();

	key = (unsigned char)(e->message & charCodeMask);
	code = (unsigned char)((e->message & keyCodeMask) >> 8);
	mods = 0;
	if (e->modifiers & cmdKey) mods |= KKeyControl;
	if (e->modifiers & shiftKey) mods |= KKeyShift;
	if (e->modifiers & optionKey) mods |= KKeyOption;
	if (e->modifiers & controlKey) mods |= KKeyCommand;
	
	/*
	 *	Determine if this is a keyboard event or a keyboard command
	 */
	
	switch (key) {
		case 0x1C:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyLeftCursor),NULL);
			break;
		case 0x1D:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyRightCursor),NULL);
			break;
		case 0x1E:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyUpCursor),NULL);
			break;
		case 0x1F:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyDownCursor),NULL);
			break;
			
		case 0x0B:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyPageUp),NULL);
			break;
		case 0x0C:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyPageDown),NULL);
			break;
		case 0x01:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyHome),NULL);
			break;
		case 0x04:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyEnd),NULL);
			break;
			
		case 0x1B:
			if (code != 0x35) f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyClear),NULL);
			else f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyEscape),NULL);
			break;

		case 0x7F:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyDelete),NULL);
			break;
		case 0x05:
			f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyHelp),NULL);
			break;

		case 0x10:						// F key
			switch (code) {
				case 0x7A:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF1),NULL);
					break;
				case 0x78:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF2),NULL);
					break;
				case 0x63:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF3),NULL);
					break;
				case 0x76:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF4),NULL);
					break;
				
				case 0x60:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF5),NULL);
					break;
				case 0x61:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF6),NULL);
					break;
				case 0x62:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF7),NULL);
					break;
				case 0x64:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF8),NULL);
					break;
				
				case 0x65:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF9),NULL);
					break;
				case 0x6D:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF10),NULL);
					break;
				case 0x67:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF11),NULL);
					break;
				case 0x6F:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF12),NULL);
					break;
				
				case 0x69:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF13),NULL);
					break;
				case 0x6B:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF14),NULL);
					break;
				case 0x71:
					f->SendDispatch(KEventKey,MAKEDWORD(mods | KSpecialKey,KKeyF15),NULL);
					break;
			}
			break;
			
		default:
			f->SendDispatch(KEventKey,MAKEDWORD(mods,key),NULL);
			break;
	}
}

/*	XGAppCore::DoOSEvent
 *
 *		Process operating system events
 */

void XGAppCore::DoOSEvent(EventRecord *e)
{
	WindowPtr w;
	short mode = (short)(0xFFL & ((e->message) >> 24));

#if OPT_FLOATWINDOW == 1
	w = MFrontWindow();
#else
	w = FrontWindow();
#endif

	if (mode == suspendResumeMessage) {
		if (e->message & resumeFlag) {
			/*
			 *	Application is being resumed from.  Activate the front window
			 */
			
			if (w != NULL) {
				HiliteWindow(w,1);
				_DoActive(w,1);
			}
			fFront = 1;
			XGBroadcast::SendBroadcast(KEventAppActivate,0,NULL);
		} else {
			/*
			 *	Application is being suspended.  Deactivate the front window
			 */
			
			if (w != NULL) {
				HiliteWindow(w,0);
				_DoActive(w,0);
			}
			fFront = 0;
			XGBroadcast::SendBroadcast(KEventAppDeactivate,0,NULL);
		}
	}
}

/************************************************************************/
/*																		*/
/*	Menu Management														*/
/*																		*/
/************************************************************************/

/*	XGAppCore::DoMenu
 *
 *		Translate the menu and dispatch it
 */

void XGAppCore::DoMenu(long msg)
{
	if (msg == 0) return;
	_GetCurMenu()->_DispatchCommand(_GetCurFocus(),GETHIWORD(msg),GETLOWORD(msg));
}

/*	XGAppCore::DoClose
 *
 *		Do the close command
 */

bool XGAppCore::DoClose(WindowPtr w)
{
	XGWindow *tw;
	int i;
	
	if (w == NULL) return true;
	i = ((WindowPeek)w)->windowKind;
	if (i < 0) {
		::CloseDeskAcc(i);			// Killed desk accessory
		return true;
	} else {
		tw = XGWindow::_GetXGWindow(w);
		if (tw) {
			if (!(tw->_Attributes() & KAttrNoClose)) return tw->CloseWindow();
			return false;
		} else {
			::CloseWindow(w);
			return true;
		}
	}
}


/************************************************************************/
/*																		*/
/*	Copyright Notice													*/
/*																		*/
/************************************************************************/

/*	XGAppCore::ShowCopyright 
 *
 *		This routine shows the copyright notice 
 */

void XGAppCore::ShowCopyright(void)
{
	ShowMacCopyright();
}
