/*	XGAppMultiWindow.cpp
 *
 *		This creates a single-window application.
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XDataUtil.h>
#include <XCursorUtil.h>
#include <XWindow.h>
#include <XMenu.h>
#include <XError.h>
#include <XApplication.h>

/************************************************************************/
/*																		*/
/*	Support routines													*/
/*																		*/
/************************************************************************/

/*	ToUpper
 *
 *		My lower case to upper case routine
 */

static char ToUpper(char c)
{
	if ((c >= 'a') && (c <= 'z')) return c + ('A' - 'a');
	else return c;
}

/*	StringCompare
 *
 *		String compare: which is bigger and which is smaller
 */

static int StringCompare(char *a, char *b)
{
	while ((*a) && (ToUpper(*a) == ToUpper(*b))) a++, b++;
	
	if (*a == *b) return 0;
	if (ToUpper(*a) < ToUpper(*b)) return -1;
	return 1;
}

/*	WindowCompare
 *
 *		This compares these two windows
 */

static int WindowCompare(XGWindow *a, XGWindow *b)
{
	char bufa[256];
	char bufb[256];
	short la,lb;
	
	la = GETLOWORD(a->_Attributes());
	lb = GETLOWORD(b->_Attributes());
	if (la < lb) return -1;
	if (la > lb) return 1;

	a->GetWindowName(bufa);
	b->GetWindowName(bufb);
	return StringCompare(bufa,bufb);
}

/*	IsWindowMenuID
 *
 *		Is this in the range of my window menu identifiers?
 */

static bool IsWindowMenuID(long id)
{
	if (id == KMenuWindowList) return 1;
	if ((id >= 0x7000) && (id < 0x8000)) return 1;
	return 0;
}

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGAppMultiWindow::XGAppMultiWindow
 *
 *		Create me
 */

XGAppMultiWindow::XGAppMultiWindow(char *appName)
{
#if OPT_MACOS == 1
	#pragma unused(appName)

	fMyMenuBar = new XGMenuBar();
#endif

#if OPT_WINOS == 1
	if (_GPrevInstance == NULL) _InitApplication();
	_InitInstance(appName);
	
	fMyMenuBar = new XGMenuBar(true,_GMDIWindow);
#endif

	fMyMenuBar->_LoadFromResource(KDefaultMenu);
}

/*	XGAppMultiWindow:::~XGAppMultiWindow
 *
 *		Delete me
 */

XGAppMultiWindow::~XGAppMultiWindow()
{
	if (fMyMenuBar) delete fMyMenuBar;
}

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGAppMultiWindow::ReceiveBroadcast
 *
 *		Process postperiodic event to update menu bar
 */

void XGAppMultiWindow::ReceiveBroadcast(long msg, long, void *)
{
	XGWindow *tw;
	
	switch (msg) {
		case KEventPostPeriodic:
			/*
			 *	Update the menu bar
			 */
			
			tw = XGWindow::GetFrontWindow();

			if (fMyMenuBar != NULL) {
				if (tw) {
					fMyMenuBar->_LoadFromResource(tw->GetMenuBarID());
				} else {
					fMyMenuBar->_LoadFromResource(KDefaultMenu);
				}
				fMyMenuBar->_UpdateMenuBar(_GetCurFocus());
			}
			break;
	}
}

/*	XGAppMultiWindow::ReceiveDispatch
 *
 *		This handles receiving various messages, and manages my application
 *	core based on the window construction/destruction messages
 */

long XGAppMultiWindow::ReceiveDispatch(long msg, long arg, void *parg)
{
	XGSMenuStatusRecord *gs;
	long len,off;
	XGWindow *w;
		
	switch (msg) {
		case KEventWindowCreate:
			_CreateWindow((XGWindow *)parg);
			break;
		case KEventWindowDestroy:
			_DestroyWindow((XGWindow *)parg);
			break;
		case KEventWindowName:
			_RenameWindow((XGWindow *)parg);
			break;
			
		case KEventGetMenuStatus:
			if (IsModalWindow()) break;				// skip if modal up
			
			gs = (XGSMenuStatusRecord *)parg;
			w = XGWindow::GetFrontWindow();
			if (IsWindowMenuID(arg) && (arg != KMenuWindowList)) {
				gs->enable = true;
				
				len = fWindowList.Length();
				for (off = 0; off < len; off++) {
					if (fWindowList[off].fWindow == w) {
						if (fWindowList[off].menuID == arg) {
							gs->checked = true;
						}
						break;
					}
				}
				return 1;
			}
			switch (arg) {
				/* Handle tile/cascade stuff */
				case KMenuWindowTile:
				case KMenuWindowCascade:
				case KMenuWindowArrange:
					if (fWindowList.Length() > 0) {
						gs->enable = true;
						return 1;
					}
					break;
			}
			break;
		case KEventDoMenuCommand:
			if (IsModalWindow()) break;				// skip if modal up
			
			if (IsWindowMenuID(arg) && (arg != KMenuWindowList)) {
				/* Find the window to enable */
				len = fWindowList.Length();
				for (off = 0; off < len; off++) {
					if (fWindowList[off].menuID == arg) {
						fWindowList[off].fWindow->SelectWindow();
						return 1;
					}
				}
			}
			switch (arg) {
				case KMenuWindowTile:
					_TileWindows();
					return 1;
				case KMenuWindowCascade:
					_CascadeWindows();
					return 1;
#if OPT_WINOS == 1
				case KMenuWindowArrange:
					_ArrangeIcons();
					return 1;
#endif
			}
			break;
	}
		
	/*
	 *	Pass event upwards
	 */
	
	return XGAppCore::ReceiveDispatch(msg,arg,parg);
}

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGAppMultiWindow::_GetCurMenu
 *
 *		The current menu for this single window application
 */

XGMenuBar *XGAppMultiWindow::_GetCurMenu()
{
	return fMyMenuBar;
}

/*	XGAppMultiWindow::_GetCurFocus
 *
 *		The current menu for this single window application
 */

XGFocus *XGAppMultiWindow::_GetCurFocus()
{
	XGWindow *tw = XGWindow::GetFrontWindow();

	if (tw) {
		return tw;					/* Front window is focus */
	} else {
		return this;				/* I'm the focus */
	}
}

/************************************************************************/
/*																		*/
/*	Window Menu Management												*/
/*																		*/
/************************************************************************/

/*	XGAppMultiWindow::_CreateWindow
 *
 *		Insert this window into the list of windows that are being
 *	managed
 */

void XGAppMultiWindow::_CreateWindow(XGWindow *w)
{
	_WLstRecord wl;
	
	/*
	 *	Screen out non-document windows
	 */
	
	if (w->GetWindowType() != KWinTypeDocument) return;
	if (w->_Attributes() & KAttrNoMenu) return;
	
	/*
	 *	Insert this thing into my list of windows
	 */
	
	wl.menuAccel = -1;
	wl.menuID = -1;
	wl.fWindow = w;
	
	fWindowList += wl;
	
	/*
	 *	Sort and update the window list
	 */
	
	_SortWindowList();
	_UpdateWindowMenu();
}

/*	XGAppMultiWindow::_DestroyWindow
 *
 *		Scan the list of windows that are in me, and if this one is,
 *	delete it
 */

void XGAppMultiWindow::_DestroyWindow(XGWindow *w)
{
	long len,off;
	
	/*
	 *	Screen out non-document windows
	 */
	
	if (w->GetWindowType() != KWinTypeDocument) return;
	if (w->_Attributes() & KAttrNoMenu) return;
	
	/*
	 *	Find window
	 */
	
	len = fWindowList.Length();
	for (off = 0; off < len; off++) {
		if (fWindowList[off].fWindow == w) break;
	}
	
	/*
	 *	Delete it
	 */
	
	if (off >= len) return;
	fWindowList.Delete(off);
	
	/*
	 *	Update the window list
	 */
	
	_UpdateWindowMenu();
}

/*	XGAppMultiWindow::_RenameWindow
 *
 *		handle the fact that this window was renamed
 */

void XGAppMultiWindow::_RenameWindow(XGWindow *w)
{
	long len,off;
	
	/*
	 *	Screen out non-document windows
	 */
	
	if (w->GetWindowType() != KWinTypeDocument) return;
	if (w->_Attributes() & KAttrNoMenu) return;
	
	/*
	 *	Find window
	 */
	
	len = fWindowList.Length();
	for (off = 0; off < len; off++) {
		if (fWindowList[off].fWindow == w) break;
	}
	
	/*
	 *	Sort and update if it was there
	 */
	
	if (off >= len) return;
	_SortWindowList();
	_UpdateWindowMenu();
}

/************************************************************************/
/*																		*/
/*	Window Menu Construction											*/
/*																		*/
/************************************************************************/

/*	XGAppMultiWindow::_SortWindowList
 *
 *		Sort this window list.
 */

void XGAppMultiWindow::_SortWindowList(void)
{
	long len,off;
	_WLstRecord wl;
	
	/*
	 *	Resort this list. This does a simple bubble sort because
	 *	I assume my window is no more than about 20 elements or so.
	 */
	
	len = fWindowList.Length();
	for (;;) {
		for (off = 1; off < len; off++) {			
			if (WindowCompare(fWindowList[off-1].fWindow,
							  fWindowList[off  ].fWindow) > 0) {
				/* Switch */
				wl = fWindowList[off-1];
				fWindowList[off-1] = fWindowList[off];
				fWindowList[off] = wl;
			}
		}
		if (off >= len) break;
	}
}

/*	XGAppMultiWindow::_UpdateWindowMenu
 *
 *		This updates the menu window (if it exists) with the new
 *	window list.
 */

void XGAppMultiWindow::_UpdateWindowMenu()
{
	long menu;
	long len;
	long off;
	long lmenu;
	long loff;
	long id;
	short accel;
	char buffer[256];
	
	/*
	 *	First, scan the menu bar until I find the menu which 
	 *	contains either my placeholder or one of my windows
	 */
	
	len = fMyMenuBar->GetNumMenus();
	for (off = 0; off < len; off++) {
		menu = fMyMenuBar->GetMenu(off);

		lmenu = fMyMenuBar->GetNumItems(menu);
		for (loff = 0; loff < lmenu; loff++) {
			id = fMyMenuBar->GetItemID(menu,loff);
			if (IsWindowMenuID(id)) break;
		}
		if (loff < lmenu) break;
	}
	if (off >= len) return;			/* Couldn't find it. */
	
	/*
	 *	if we get here, then
	 *
	 *		menu			Is the reference to the menu containing my
	 *						window menu list
	 *		loff			Is the first item which refers to my menu
	 *
	 *	Next, we delete what was there before
	 */
	
	for (;;) {
		lmenu = fMyMenuBar->GetNumItems(menu);
		if (loff >= lmenu) break;
		id = fMyMenuBar->GetItemID(menu,loff);
		if (!IsWindowMenuID(id)) break;
		
		fMyMenuBar->DeleteMenuItem(menu,loff);
	}
	
	/*
	 *	Update the accelerators in my menu list to assign those
	 *	where I have a free accelerator. (We only assign accelerator
	 *	keys '1' through '9'; we expect '0' is reserved.)
	 */
	
	len = fWindowList.Length();
	for (off = 0; off < len; off++) {
		if (fWindowList[off].menuAccel == -1) {
			/* No accelerator; scan for one */
			for (accel = 1; accel <= 9; accel++) {
				for (id = 0; id < len; id++) {
					if (fWindowList[id].menuAccel == accel) break;
				}
				if (id >= len) break;
			}
			if (accel >= 10) break;
			
			fWindowList[off].menuAccel = accel;
		}
	}
	
	/*
	 *	Update the menu identifier to assign an empty identifier
	 */
	
	for (off = 0; off < len; off++) {
		if (fWindowList[off].menuID == -1) {
			/* No accelerator; scan for one */
			for (accel = 0x7000; accel < 0x8000; accel++) {
				for (id = 0; id < len; id++) {
					if (fWindowList[id].menuID == accel) break;
				}
				if (id >= len) break;
			}
			if (accel >= 0x8000) break;
			
			fWindowList[off].menuID = accel;
		}
	}
	
	/*
	 *	Now insert the list into the menu bar
	 */
	
	if (len == 0) {
		fMyMenuBar->InsertMenuItem(menu,loff);
		fMyMenuBar->SetItemID(menu,loff,KMenuWindowList);
		fMyMenuBar->SetItemText(menu,loff,"Empty");
	} else {
		for (off = 0; off < len; off++, loff++) {
			fMyMenuBar->InsertMenuItem(menu,loff);
			fMyMenuBar->SetItemID(menu,loff,fWindowList[off].menuID);
			
#if OPT_MACOS == 1
			fWindowList[off].fWindow->GetWindowName(buffer);
			fMyMenuBar->SetItemText(menu,loff,buffer);
			if (fWindowList[off].menuAccel != -1) {
				fMyMenuBar->SetItemShortcut(menu,loff,'0' + fWindowList[off].menuAccel);
			}
#endif

#if OPT_WINOS == 1
			if (fWindowList[off].menuAccel == -1) {
				fWindowList[off].fWindow->GetWindowName(buffer);
			} else {
				buffer[0] = '&';
				buffer[1] = '0' + fWindowList[off].menuAccel;
				buffer[2] = ' ';
				fWindowList[off].fWindow->GetWindowName(buffer+3);
			}
			fMyMenuBar->SetItemText(menu,loff,buffer);
#endif

		}
	}
}

/************************************************************************/
/*																		*/
/*	Tiling/Cascading Windows (Macintosh Version)						*/
/*																		*/
/************************************************************************/

#if OPT_MACOS == 1

/*	XGAppMultiWindow::_TileWindows
 *
 *		Tile the windows across the screen
 */
 
void XGAppMultiWindow::_TileWindows()
{
	Rect r,s,tmp;
	short count;
	short i,mid;
	Point nsize;
	XGWindow *foo;
	
	count = fWindowList.Length();
	if (count == 0) return;				// Huh?  Something didn't compute
	XGWindow::GetScreenSize(&r);		// Get bounding of the screen
	
	if (count == 1) {
		InsetRect(&r,2,2);
		
		foo = fWindowList[0].fWindow;
		nsize.h = r.right - r.left;
		nsize.v = r.bottom - r.top;
		foo->_LimitWindowSize(&nsize);
		r.right = r.left + nsize.h;
		r.bottom = r.top + nsize.v;
		foo->SetWindowLocation(r);
	} else if (count == 2) {
		s = r;
		s.bottom = (r.top + r.bottom)/2;
		InsetRect(&s,1,1);
		foo = fWindowList[0].fWindow;
		nsize.h = s.right - s.left;
		nsize.v = s.bottom - s.top;
		foo->_LimitWindowSize(&nsize);
		s.right = s.left + nsize.h;
		s.bottom = s.top + nsize.v;
		foo->SetWindowLocation(s);
		
		s = r;
		s.top = (r.top + r.bottom)/2;
		InsetRect(&s,1,1);
		foo = fWindowList[1].fWindow;
		nsize.h = s.right - s.left;
		nsize.v = s.bottom - s.top;
		foo->_LimitWindowSize(&nsize);
		s.right = s.left + nsize.h;
		s.bottom = s.top + nsize.v;
		foo->SetWindowLocation(s);
	} else {
		mid = count/2;
		s = r;
		s.right = (r.right+r.left)/2;
		for (i = 0; i < mid; i++) {
			WaitCursor();
			tmp = s;
			tmp.top = ((r.bottom-r.top)*i)/mid + s.top;
			tmp.bottom = ((r.bottom-r.top)*(i+1))/mid + s.top;
			InsetRect(&tmp,1,1);
			
			foo = fWindowList[i].fWindow;
			nsize.h = tmp.right - tmp.left;
			nsize.v = tmp.bottom - tmp.top;
			foo->_LimitWindowSize(&nsize);
			tmp.right = tmp.left + nsize.h;
			tmp.bottom = tmp.top + nsize.v;
			foo->SetWindowLocation(tmp);
		}
		s = r;
		s.left = (r.right+r.left)/2;
		for (i = mid; i < count; i++) {
			WaitCursor();
			tmp = s;
			tmp.top = ((r.bottom-r.top)*(i-mid))/(count-mid) + s.top;
			tmp.bottom = ((r.bottom-r.top)*(i-mid+1))/(count-mid) + s.top;
			InsetRect(&tmp,1,1);

			foo = fWindowList[i].fWindow;
			nsize.h = tmp.right - tmp.left;
			nsize.v = tmp.bottom - tmp.top;
			foo->_LimitWindowSize(&nsize);
			tmp.right = tmp.left + nsize.h;
			tmp.bottom = tmp.top + nsize.v;
			foo->SetWindowLocation(tmp);
		}
		NormalCursor();
	}
}

/*	XGAppMultiWindow::_CascadeWindows
 *
 *		Cascade the windows on the screen
 */

void XGAppMultiWindow::_CascadeWindows()
{
	Rect r,s;
	XGWindow *foo;
	int count;
	Point size;
	Point nsize;
	int i;
	
	count = fWindowList.Length();
	if (count == 0) return;
	XGWindow::GetScreenSize(&r);
	r.bottom -= 2;
	
	if (count > 15) count = 13;
	size.h = (r.right - r.left - 10) - 5 * (count - 1);
	if (count > 10) count = 10;
	size.v = (r.bottom - r.top) - 20 * (count - 1);
	
	count = fWindowList.Length();
	for (i = 0; i < count; i++) {
		WaitCursor();
		
		nsize = size;
		foo = fWindowList[i].fWindow;
		foo->_LimitWindowSize(&nsize);

		s.left = (i % 13) * 5 + r.left + 5;
		s.right = s.left + nsize.h;
		s.top = (i % 10) * 20 + r.top;
		s.bottom = s.top + nsize.v;
		
		foo->SetWindowLocation(s);
	}
	NormalCursor();
}

#endif

/************************************************************************/
/*																		*/
/*	Tiling/Cascading Windows (Microsoft Windows Version)				*/
/*																		*/
/************************************************************************/

#if OPT_WINOS == 1

/*	XGAppMultiWindow::_TileWindows
 *
 *		Tile the windows across the screen
 */
 
void XGAppMultiWindow::_TileWindows()
{
	::SendMessage(_GMDIWindow,WM_MDITILE,(WPARAM)MDITILE_VERTICAL,NULL);
}

/*	XGAppMultiWindow::_CascadeWindows
 *
 *		Cascade the windows on the screen
 */

void XGAppMultiWindow::_CascadeWindows()
{
	::SendMessage(_GMDIWindow,WM_MDICASCADE,0,NULL);
}

/*	XGAppMultiWindow::_ArrangeWindows
 *
 *		Arrange the window icons at the bottom of the MDI window
 */

void XGAppMultiWindow::_ArrangeIcons()
{
	::SendMessage(_GMDIWindow,WM_MDIICONARRANGE,0,NULL);
}

#endif

/************************************************************************/
/*																		*/
/*	Windows initialization support										*/
/*																		*/
/************************************************************************/

#if OPT_WINOS == 1

/*	XGAppMultiWindow::_InitApplication
 *
 *		Register the classes which are in me
 */

void XGAppMultiWindow::_InitApplication(void)
{
	WNDCLASS wc;
	
	/*
	 *	Initialize four instances: the main window which encapsulates
	 *	the MDI window, the child window, the dialog window and the 
	 *	view window
	 */
	
	wc.style = 0;
	wc.lpfnWndProc = _GMainMDIWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = _GInstance;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_APPWORKSPACE+1);
	wc.lpszMenuName = MAKEINTRESOURCE(KDefaultMenu);
	wc.lpszClassName = _GMainClass;
	if (!RegisterClass(&wc)) {
		throw XPostError("Unable to register ^S",_GMainClass);
	}
	
	wc.style = CS_DBLCLKS;
	wc.lpfnWndProc = _GChildWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = _GInstance;
	wc.hIcon = NULL;
	wc.hCursor = NULL;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _GWindowClass;
	if (!RegisterClass(&wc)) {
		throw XPostError("Unable to register ^S",_GWindowClass);
	}
	
	wc.style = CS_DBLCLKS;
	wc.lpfnWndProc = _GViewWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = _GInstance;
	wc.hIcon = NULL;
	wc.hCursor = NULL;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _GViewClass;
	if (!RegisterClass(&wc)) {
		throw XPostError("Unable to register ^S",_GViewClass);
	}

	wc.style = CS_DBLCLKS;
	wc.lpfnWndProc = _GSingleWindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = _GInstance;
	wc.hIcon = NULL;
	wc.hCursor = NULL;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = _GDlogClass;
	if (!RegisterClass(&wc)) {
		throw XPostError("Unable to register ^S",_GDlogClass);
	}	
}

/*	XGAppMultiWindow::_InitInstance
 *
 *		Initialize the instance of this thing. This builds the MDI frame
 *	of my window.
 */

void XGAppMultiWindow::_InitInstance(char *appName)
{
	if (appName == NULL) appName = "YAAF Application";
	
	_GMainWindow = CreateWindow(_GMainClass,appName,
					WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_MAXIMIZEBOX |
					WS_MINIMIZEBOX,
					CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
					NULL, NULL, _GInstance, NULL);
	if (!_GMainWindow) throw XPostError("Unable to initialize main window");
	
	ShowWindow(_GMainWindow,_GArgShow);
	UpdateWindow(_GMainWindow);
}


#endif
