/*	XGBroadcast.cpp
 *
 *		This handles broadcasting global messages.
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XEvent.h>
#include <XThread.h>

/************************************************************************/
/*																		*/
/*	Globals																*/
/*																		*/
/************************************************************************/

static XGSemaphore GMutexPost;				// Serialize access to post stack
static XGSemaphore GMutex;					// Serialize creation/destruction

XGBroadcast *XGBroadcast::gList = NULL;
XGDynArray<_PostRecord> XGBroadcast::gPostStack;

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGBroadcast::XGBroadcast
 *
 *		Insert me into the linked list
 */

XGBroadcast::XGBroadcast()
{
	XGCritical lock(&GMutex);				// Lock entire function 
	
	fNext = gList;
	gList = this;
}

/*	XGBroadcast::~XGBroadcast
 *
 *		Remove me from the linked list
 */

XGBroadcast::~XGBroadcast()
{
	XGCritical lock(&GMutex);				// Lock entire function 
	XGBroadcast *foo;
	
	if (gList == this) gList = fNext;
	else {
		for (foo = gList; foo != NULL; foo = foo->fNext) {
			if (foo->fNext == this) {
				foo->fNext = fNext;
				break;
			}
		}
	}
}

/************************************************************************/
/*																		*/
/*	Broadcast messages													*/
/*																		*/
/************************************************************************/

/*	XGBroadcast::ReceiveBroadcast
 *
 *		Recieve this message default: do nothing
 */

void XGBroadcast::ReceiveBroadcast(long, long, void *)
{
}

/*	XGBroadcast::SendBroadcast
 *
 *		Broadcast this message globally
 */

void XGBroadcast::SendBroadcast(long msg, long arg, void *parg)
{
	XGBroadcast *foo;
	
	for (foo = gList; foo != NULL; foo = foo->fNext) {
		foo->ReceiveBroadcast(msg,arg,parg);
	}
}

/*	XGBroadcast::PostBroadcast
 *
 *		This posts this message to be broadcast later. This is useful
 *	in a multi-threaded application where we wish to have delayed
 *	notification when things are relatively safe again. (For example,
 *	change focus commands are broadcast instead of posted.)
 */

void XGBroadcast::PostBroadcast(long msg, long arg, void *parg)
{
	XGCritical lock(&GMutexPost);			// Lock entire function
	_PostRecord pr;
	
	pr.msg = msg;
	pr.arg = arg;
	pr.parg = parg;
	gPostStack += pr;
}

/*	XGBroadcast::_SendPost
 *
 *		Send the posted stuff. This sends posts (one at a time, peeled
 *	from the bottom) until they are exausted.
 */

void XGBroadcast::_SendPost(void)
{
	_PostRecord p;
	
	while (gPostStack.Length()) {
		GMutexPost.EnterCritical();
		p = gPostStack[0];
		gPostStack.Delete(0);			// delete the bottom of the stack
		GMutexPost.LeaveCritical();

		XGBroadcast::SendBroadcast(p.msg,p.arg,p.parg);
	}
}
