/*	XGDispatch.cpp
 *
 *		The dispatcher routine
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XThread.h>
#include <XEvent.h>

/************************************************************************/
/*																		*/
/*	Globals																*/
/*																		*/
/************************************************************************/

static XGSemaphore GMutex;

XGDispatch *XGDispatch::gDefault = NULL;
XGDispatch *XGDispatch::gParent = NULL;

/************************************************************************/
/*																		*/
/*	Constructor/Destructor												*/
/*																		*/
/************************************************************************/

/*	XGDispatch::XGDispatch
 *
 *		Handle dispatch construction. This is a bit wierd, but it's done
 *	for a couple of reasons.
 *
 *		First, I assume the XGApplication class, which is the root class
 *	in my world, is the first XGDispatch object to be constructed. Thus,
 *	we can construct that class without any arguments, and it will set
 *	me as the top of the event tree.
 *
 *		Once we have the top of the tree, we can now create classes whose
 *	parents are the tree top (such as lone windows and documents) without
 *	having to know who is the current tree top.
 */

XGDispatch::XGDispatch(void)
{
	XGCritical lock(&GMutex);		// Lock entire function 
	
	if (gParent == NULL) {
		fParent = gDefault;
		if (gDefault == NULL) {
			gDefault = this;
		}
	} else {
		fParent = gParent;
		fFocus = gParent->fFocus;	// By default I'm in my parent's focus area
	}
	
	gParent = this;
}

/*	XGDispatch::~XGDispatch
 *
 *		Destroy me. This simply checks to make sure I wasn't the focus.
 */

XGDispatch::~XGDispatch()
{
	XGCritical lock(&GMutex);		// Lock entire function 
	
	if (fFocus->GetFocus() == this) {
		fFocus->_ClearFocus();		// Move focus to the root object for this
	}
}

/************************************************************************/
/*																		*/
/*	Focus Management													*/
/*																		*/
/************************************************************************/

/*	XGDispatch::GainFocus
 *
 *		Default behaviour does nothing
 */

void XGDispatch::GainFocus()
{
}

/*	XGDispatch::LoseFocus
 *
 *		Default behaviour does nothing
 */

void XGDispatch::LoseFocus()
{
}

/*	XGDispatch::HasFocus
 *
 *		Do I have focus?
 */

bool XGDispatch::HasFocus(void)
{
	if (fFocus->GetFocus() == this) return true;
	return false;
}

/*	XGDispatch::SetFocus
 *
 *		Set the focus. This broadcasts a 'changed focus' message globally
 *	and changes the focus to me
 */

XGDispatch *XGDispatch::SetFocus()
{
	XGDispatch *tmp;
	
	if (this == fFocus->GetFocus()) return this;	// Do nothing
	
	/*
	 *	Serialize access to the SetFocus function
	 */
	
	XGCritical lock(&GMutex);						// Lock entire function 
	
	/*
	 *	Change the focus
	 */
	
	tmp = fFocus->GetFocus();
	fFocus->_SetFocus(this);
	
	/*
	 *	Now send focus change messages
	 */
	
	if (tmp) tmp->LoseFocus();
	GainFocus();
	
	/* 
	 *	And return the prior focus holder
	 */
	
	return tmp;
}


/************************************************************************/
/*																		*/
/*	Message Dispatch													*/
/*																		*/
/************************************************************************/

/*	XGDispatch:ReceiveDispatch
 *
 *		This is the default thing to call if the dispatch isn't handled
 *	above me. This will send the message to my parent
 */

long XGDispatch::ReceiveDispatch(long msg, long arg, void *parg)
{
	if (fParent) {
		return fParent->ReceiveDispatch(msg,arg,parg);
	} else {
		return 0;
	}
}
