/*	XGThread.cpp
 *
 *		Thread support. This allows me to encapsulate simple thread
 *	semantics behind a simple object declaration. And this allows me
 *	to use the same semantics on both Windows, the MacOS and MkLinux
 *	X Windows.
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XConfig.h>
#include <XError.h>
#include <XThread.h>

/************************************************************************/
/*																		*/
/*	Globals																*/
/*																		*/
/************************************************************************/

static XGSemaphore GMutex;
XGThread *XGThread::gList = NULL;

static XGThread GAppThread;			/* Application thread object		*/

#if OPT_WINOS == 1
extern DWORD GAppThreadID = -1;
#endif

/************************************************************************/
/*																		*/
/*	Thread Support														*/
/*																		*/
/************************************************************************/

/*	XGThread::XGThread
 *
 *		Create this thread.
 */

XGThread::XGThread(XGThreadProcPtr proc, long stack, bool suspend)
{
#if OPT_MACOS == 1
	/*
	 *	Mac version
	 */

	OSErr err;
	
	/*
	 *	Create this thread. The thread is built suspended so I can
	 *	have a chance to do some initialization before it runs.
	 */
	
	fThreadID = kNoThreadID;
	fProc = proc;
	fAttach = 1;					// One owner--me
	if (0 != (err = NewThread(kCooperativeThread,	// only type on MacOS
							  ThreadEntry,			// my entry point
							  (void *)this,			// and my argument to entry
							  stack,				// stack size default
							  kNewSuspend,			// suspended at first
							  NULL,					// Thread result--who cares?
							  &fThreadID)))			// Thread ID of me
		throw XPostError("Unable to create thread");
	
	/*
	 *	Now, initialize fields in this thread
	 */
	
	SetThreadTerminator(fThreadID,ThreadTerminate,(void *)this);
#endif

#if OPT_WINOS == 1
	/*
	 *	Windows 95/Windows NT version: what's the default thread ID?
	 */
	
	if (GAppThreadID == -1) {
		GAppThreadID = GetCurrentThreadId();
	}
	
	/*
	 *	Create this thread. The thread is built suspended so I can
	 *	have a chance to do some initialization before it runs
	 */
	
	fProc = proc;
	fAttach = 1;
	fThreadHandle = CreateThread(NULL,				// No security params
								 stack,				// default stack size
								 ThreadEntry,		// my entry point
								 (LPVOID)this,		// and my argument
								 CREATE_SUSPENDED,	// create suspended
								 &fThreadID);		// and what am I?
	if (fThreadHandle == NULL) {
		throw XPostError("Unable to create thread");
	}
	
	fThreadSuspend = true;
#endif
		
	/*
	 *	Insert thread into linked list
	 */
	
	GMutex.EnterCritical();
	fNext = gList;
	gList = this;
	GMutex.LeaveCritical();
	
	/*
	 *	And resume the thread if it's what the user wants
	 */
	
	if (!suspend) Resume();
}

/*	XGThread::XGThread(void)
 *
 *		This is a special constructor, which should never be called by
 *	another procedure, which creates the thread which represents the main
 *	task.
 */

XGThread::XGThread()
{
	GMutex.EnterCritical();
	
#if OPT_MACOS == 1
	fThreadID = kApplicationThreadID;
#endif

#if OPT_WINOS == 1
	if (GAppThreadID == -1) GAppThreadID = GetCurrentThreadId();
	fThreadID = GAppThreadID;
	fThreadHandle = NULL;
#endif

	fProc = NULL;
	fNext = gList;
	gList = this;
	
	GMutex.LeaveCritical();
}

/*	XGThread::~XGThread
 *
 *		Stop this thing if it's running
 */

XGThread::~XGThread()
{
#if OPT_MACOS == 1
	ThreadID t;
	
	if ((fThreadID != kApplicationThreadID) && (fThreadID != kNoThreadID)) {
		/*
		 *	We've got to kill this first
		 */
		
		t = fThreadID;
		fThreadID = kNoThreadID;
		DisposeThread(t,NULL,false);
	}
#endif

#if OPT_WINOS == 1
	if (fThreadHandle) Kill();
#endif
	
	/*
	 *	Remove thread from linked list
	 */
	
	XGThread *thread;

	GMutex.EnterCritical();
	if (gList == this) gList = fNext;
	else {
		for (thread = gList; thread != NULL; thread = thread->fNext) {
			if (thread->fNext == this) {
				thread->fNext = fNext;
				break;
			}
		}
	}
	GMutex.LeaveCritical();
}

/************************************************************************/
/*																		*/
/*	Thread Management													*/
/*																		*/
/************************************************************************/

/*	XGThread::YieldThread
 *
 *		Yield to any thread
 */

void XGThread::YieldThread(bool f)
{
#if OPT_MACOS == 1
	#pragma unused(f)
	YieldToAnyThread();			/* Always yield computational time */
#endif

#if OPT_WINOS == 1
	if (f) Sleep(0);			/* Yield rest of my timeslice */
#endif
}

/*	XGThread::CurrentThread
 *
 *		Which is the current thread object? Get the thread ID and search
 *	for the current thread object
 */

XGThread *XGThread::CurrentThread()
{
	XGThread *t;

#if OPT_MACOS == 1
	ThreadID id;
	
	if (GetCurrentThread(&id)) return &GAppThread;
	if (id == kApplicationThreadID) return &GAppThread;
#endif

#if OPT_WINOS == 1
	DWORD id;
	
	id = GetCurrentThreadId();
	if (id == GAppThreadID) return &GAppThread;
#endif

	GMutex.EnterCritical();
	for (t = gList; t != NULL; t = t->fNext) {
		if (t->fThreadID == id) break;
	}
	GMutex.LeaveCritical();
	
	if (t == NULL) return &GAppThread;
	return t;
}

/************************************************************************/
/*																		*/
/*	Thread Management													*/
/*																		*/
/************************************************************************/

/*	XGThread::Suspend
 *
 *		Suspend the specified thread
 */

void XGThread::Suspend(void)
{
#if OPT_MACOS == 1
	if (fThreadID != kNoThreadID) {
		::SetThreadState(fThreadID,kStoppedThreadState,kNoThreadID);
	}
#endif

#if OPT_WINOS == 1
	if (!IsSuspend() && (fThreadHandle != NULL)) {
		::SuspendThread(fThreadHandle);
		fThreadSuspend = false;
	}
#endif
}

/*	XGThread::Resume
 *
 *		Resume specified thread
 */

void XGThread::Resume(void)
{
#if OPT_MACOS == 1
	if (fThreadID != kNoThreadID) {
		SetThreadState(fThreadID,kReadyThreadState,kNoThreadID);
	}
#endif

#if OPT_WINOS == 1
	if (IsSuspend() && (fThreadHandle != NULL)) {
		::ResumeThread(fThreadHandle);
		fThreadSuspend = true;
	}
#endif
}

/*	XGThread::IsSuspend
 *
 *		Is this thread suspended?
 */

bool XGThread::IsSuspend(void)
{
#if OPT_MACOS == 1
	ThreadState state;
	
	if (fThreadID == kNoThreadID) return true;		/* Like it's not running */
	GetThreadState(fThreadID,&state);
	return (state == kStoppedThreadState) ? true : false;
#endif

#if OPT_WINOS == 1
	return fThreadSuspend;
#endif
}

/*	XGThread::Kill
 *
 *		Kill this thread
 */

void XGThread::Kill(void)
{
#if OPT_MACOS == 1
	ThreadID t;
	
	if (fThreadID == kNoThreadID) return;
	if (fThreadID == kApplicationThreadID) {
		throw XPostError("Unable to kill main application thread.");
	}
	
	t = fThreadID;
	fThreadID = kNoThreadID;
	DisposeThread(t,NULL,false);
#endif

#if OPT_WINOS == 1
	if (fThreadHandle) {
		TerminateThread(fThreadHandle,0);
		fThreadHandle = NULL;
		Detach();
	}
#endif
}

/*	XGThread::IsAlive
 *
 *		Is this thread alive?
 */

bool XGThread::IsAlive(void)
{
#if OPT_MACOS == 1
	return (fThreadID != NULL) ? true : false;
#endif

#if OPT_WINOS == 1
	if (fThreadID == GAppThreadID) return true;
	if (fThreadHandle != NULL) return true;
	return false;
#endif
}

/************************************************************************/
/*																		*/
/*	Internal thread utilities											*/
/*																		*/
/************************************************************************/

#if OPT_MACOS == 1

/*	XGThread::ThreadEntry
 *
 *		This is the actual thread entry point. This wrappers some stuff
 *	for my thread. This contains the throw/catch routines to prevent
 *	tossing this exception across the stack boundary.
 */

pascal void *XGThread::ThreadEntry(void *ptr)
{
	XGThread *tptr = (XGThread *)ptr;
	
	try {
		if (tptr->fProc) (tptr->fProc)(tptr);
	}
	catch (XError &x) {
		// If it's one of mine, do the default error display
		x.DisplayError();
	}
	catch (...) {
		// all the rest get eaten and absorbed.
		XPostError te("Random exception caught by XGThread::ThreadEntry");
		te.DisplayError();
	}
	
	return NULL;
}

/*	XGThread::ThreadTerminate
 *
 *		This is the termination routine; this handles detaching this
 *	thread object
 */

pascal void XGThread::ThreadTerminate(ThreadID,void *tptr)
{
	XGThread *t = (XGThread *)tptr;
	
	t->Detach();
}

#endif

#if OPT_WINOS == 1

/*	XGThread::ThreadEntry
 *
 *		The entrypoint for the Windows application
 */

DWORD WINAPI XGThread::ThreadEntry(LPVOID ptr)
{
	XGThread *tptr = (XGThread *)ptr;
	
	try {
		if (tptr->fProc) (tptr->fProc)(tptr);
	}
	catch (XError &x) {
		// If it's one of mine, do the default error display
		x.DisplayError();
	}
	catch (...) {
		// all the rest get eaten and absorbed.
		XPostError te("Random exception caught by XGThread::ThreadEntry");
		te.DisplayError();
	}

	tptr->fThreadHandle = NULL;
	tptr->Detach();
	return 0;
}

#endif

