/*	XGDraw.cpp
 *
 *		This is the stack object which provides proper serial access
 *	to drawing stuff
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XError.h>
#include <XThread.h>
#include <XView.h>
#include <XPrinter.h>
#include <XWindow.h>

/************************************************************************/
/*																		*/
/*	Mutex Objects														*/
/*																		*/
/************************************************************************/

static XGThread	*GAccess;
static XGSemaphore GMutex;

XGDraw *XGDraw::gDraw = NULL;

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGDraw::XGDraw
 *
 *		This is a bit tricky.
 *
 *		In order to provide proper serialization to drawing, yet provide
 *	nested drawing (where one drawing routine may call the drawing
 *	routine in another view), this provides a mechanism for allowing
 *	only one thread access, yet allowing nesting of these drawing objects.
 *
 *		This has a couple of consequences.
 *
 *		First, the XGDraw objects must be released fairly quick; simply
 *	grabbing onto the XGDraw object for the duration of the application
 *	is a problem.
 *
 *		Second, firing off a second thread to draw within a draw object
 *	will block until the first draw object exits.
 */

XGDraw::XGDraw(XGView *view, bool lc)
{
	XGThread *thread;
	
	/*
	 *	Spin lock until the drawing stack is free, or if this
	 *	is my thread
	 */
	
	thread = XGThread::CurrentThread();	// What thread is up?
	
	GMutex.EnterCritical();				// Can only test one thread at a time
	for (;;) {
		if (GAccess == thread) break;
		if (gDraw == NULL) break;
		
		// Conditions not met: this thread cannot access.
		GMutex.LeaveCritical();			// Release my creator lock
		XGThread::YieldThread(true);	// Yield to someone else
		GMutex.EnterCritical();			// Enter my creator lock
	}
	
	/*
	 *	This thread is granted drawing access. Initialize this
	 *	drawing state object
	 */
	
	GAccess = thread;					// Note who I am
	fLast = gDraw;						// and insert me into the linked list
	gDraw = this;
	fView = view;
	fPrint = NULL;

	if (fLast) fLast->SaveContext();	// Save old drawing context
	CreateContext(lc);
	
	/*
	 *	And release my mutex. When we leave this thing, we are the
	 *	one thread that is allowed drawing access--only threads
	 *	that I want to have drawing access can ever leave
	 */
	 
	GMutex.LeaveCritical();				// Leave critical access
}

XGDraw::XGDraw(XGPrinter *print)
{
	XGThread *thread;
	
	/*
	 *	Spin lock until the drawing stack is free, or if this
	 *	is my thread
	 */
	
	thread = XGThread::CurrentThread();	// What thread is up?
	
	GMutex.EnterCritical();				// Can only test one thread at a time
	for (;;) {
		if (GAccess == thread) break;
		if (gDraw == NULL) break;
		
		// Conditions not met: this thread cannot access.
		GMutex.LeaveCritical();			// Release my creator lock
		XGThread::YieldThread(true);	// Yield to someone else
		GMutex.EnterCritical();			// Enter my creator lock
	}
	
	/*
	 *	This thread is granted drawing access. Initialize this
	 *	drawing state object
	 */
	
	GAccess = thread;					// Note who I am
	fLast = gDraw;						// and insert me into linked list
	gDraw = this;
	fView = NULL;
	fPrint = print;

	if (fLast) fLast->SaveContext();	// Save old drawing context
	CreateContext(true);				// parameter is ignored (for now)
	
	/*
	 *	And release my mutex. When we leave this thing, we are the
	 *	one thread that is allowed drawing access--only threads
	 *	that I want to have drawing access can ever leave
	 */
	 
	GMutex.LeaveCritical();				// Leave critical access
}

/*	XGDraw::~XGDraw
 *
 *		Destroy this, and restore drawing to the previous draw object
 */

XGDraw::~XGDraw()
{
	GMutex.EnterCritical();				// Serialize leaving this too
	
	DestroyContext();					// Release the current drawing context
	gDraw = fLast;
	if (gDraw) gDraw->RestoreContext();	// Restore last drawing context
	
	GMutex.LeaveCritical();
}

/************************************************************************/
/*																		*/
/*	Internal Support													*/
/*																		*/
/************************************************************************/

/*	XGDraw::ClipRect
 *
 *		Clip inside this rectangle. Allow the entire area if NULL. Does
 *	not work if the 'lc' flag is clear--like I care.
 */

void XGDraw::ClipRect(Rect *s)
{
#if OPT_MACOS == 1
	Rect r;
	
	r = fView->fGlobalRect;
	::OffsetRect(&r,-fView->fGlobalPos.h,-fView->fGlobalPos.v);
	if (s) SectRect(s,&r,&r);

	::ClipRect(&r);
#endif

#if OPT_WINOS == 1
	HRGN rgn;

	if (s == NULL) ::SelectClipRgn(fDC,NULL);
	else {	
		rgn = ::CreateRectRgnIndirect(s);
		::SelectClipRgn(fDC,rgn);
		::DeleteObject(rgn);
	}
#endif
}

/************************************************************************/
/*																		*/
/*	Drawing Context	(Macintosh)											*/
/*																		*/
/************************************************************************/

#if OPT_MACOS == 1

/*	XGDraw::SaveContext
 *
 *		Save the drawing context
 */

void XGDraw::SaveContext()
{
	// Save drawing port
	::GetPort(&fPort);
	
	// Save origin
	fOffset.h = fPort->portRect.left;
	fOffset.v = fPort->portRect.top;
	
	// Save clipping rectangle
	fClip = ::NewRgn();
	::GetClip(fClip);
	
	// Save pen state and color information
	::GetPenState(&fPenState);
	fFont = fPort->txFont;
	fFace = fPort->txFace;
	fSize = fPort->txSize;
	fMode = fPort->txMode;
	fSpExtra = fPort->spExtra;	
	::GetForeColor(&fForeColor);
	::GetBackColor(&fBackColor);
}

/*	XGDraw::RestoreContext
 *
 *		Restore the drawing context
 */

void XGDraw::RestoreContext()
{
	// Restore drawing port
	::SetPort(fPort);
	
	// Restore origin
	::SetOrigin(fOffset.h,fOffset.v);
	
	// Restore clipping
	::SetClip(fClip);
	::DisposeRgn(fClip);
	
	// Restore pen state and color information
	::SetPenState(&fPenState);
	::TextFont(fFont);
	::TextFace(fFace);
	::TextSize(fSize);
	::TextMode(fMode);
	::SpaceExtra(fSpExtra);
	::RGBForeColor(&fForeColor);
	::RGBBackColor(&fBackColor);
}

/*	XGDraw::CreateContext
 *
 *		Create drawing context into the geometry specified. The 'local
 *	coordinate' flag indicates that we should clip and offset the viewport
 *	so that (0,0) is the upper left of the view. This is the default.
 *
 *		The only time when we don't want to do this is when we are
 *	dealing with standard Maintosh controls.
 */

void XGDraw::CreateContext(bool lc)
{
	GrafPtr port;
	RGBColor color;
	PenState state;
	Rect r;
	
	if (fView) {
		// Set the drawing port
		port = (fView->GetWindow())->_GetGrafPtr();
		::SetPort(port);

		if (lc) {
			// Set the coordinate system so it's view local
			::SetOrigin(-fView->fGlobalPos.h,-fView->fGlobalPos.v);
			r = fView->fGlobalRect;
			OffsetRect(&r,-fView->fGlobalPos.h,-fView->fGlobalPos.v);
			::ClipRect(&r);
		} else {
			// Set the coordinate system so it's global to the window.
			// useful for Macintosh controls
			::SetOrigin(0,0);
			r = fView->fGlobalRect;
			::ClipRect(&r);
		}
	} else {
		// Set the paint port
		port = (GrafPtr)(fPrint->fPort);
		::SetPort(port);
	}
	
	// Now, set the font, text, color and other information
	color.red = 0;
	color.green = 0;
	color.blue = 0;
	::RGBForeColor(&color);
	color.red = 0xFFFF;
	color.green = 0xFFFF;
	color.blue = 0xFFFF;
	::RGBBackColor(&color);
	
	::SpaceExtra(0x10000);
	
	::TextFont(1);
	::TextSize(9);
	::TextFace(0);
	::TextMode(srcCopy);
	
	state.pnLoc.h = 0;
	state.pnLoc.v = 0;
	state.pnSize.h = 1;
	state.pnSize.v = 1;
	state.pnMode = srcCopy;
	state.pnPat = qd.black;
	::SetPenState(&state);
}

/*	XGDraw::DestroyContext
 *
 *		Destroy this context. This doesn't do a thing
 */

void XGDraw::DestroyContext()
{
}

#endif

/************************************************************************/
/*																		*/
/*	Drawing Context	(Windows)											*/
/*																		*/
/************************************************************************/

#if OPT_WINOS == 1

/*	XGDraw::SaveContext
 *
 *		Save the drawing context
 */

void XGDraw::SaveContext()
{
	fDCSave = ::SaveDC(fDC);
}

/*	XGDraw::RestoreContext
 *
 *		Restore the drawing context
 */

void XGDraw::RestoreContext()
{
	::RestoreDC(fDC,fDCSave);
}

/*	XGDraw::CreateContext
 *
 *		Create drawing context into the geometry specified. The 'local
 *	coordinate' parameter is ignored--as we are always in local coordinates
 *	with child windows.
 */

void XGDraw::CreateContext(bool)
{
	if (fView) {
		fDC = ::GetDC(fView->_GetHWND());
	} else if (fPrint) {
		fDC = fPrint->fPrintDC;
	}
}

/*	XGDraw::DestroyContext
 *
 *		Destroy this context. This doesn't do a thing
 */

void XGDraw::DestroyContext()
{
	if (fView) {
		::ReleaseDC(fView->_GetHWND(),fDC);
	} else if (fPrint) {
		// On Windows, do nothing.
	}
}

#endif
