/*	XGViewFactory.cpp
 *
 *		The view construction factory itself
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XDataUtil.h>
#include <XView.h>
#include <XError.h>
#include <XThread.h>

/************************************************************************/
/*																		*/
/*	View Factory														*/
/*																		*/
/************************************************************************/

static XGSemaphore			GMutex;

XGViewFactory *XGViewFactory::gList = NULL;

/************************************************************************/
/*																		*/
/*	View Factory														*/
/*																		*/
/************************************************************************/

/*	XGViewFactory::XGViewFactory
 *
 *		Construct me
 */

XGViewFactory::XGViewFactory()
{
	XGCritical lock(&GMutex);				// Lock entire function 
	
	fNext = gList;
	gList = this;
}

/*	XGViewFactory::~XGViewFactory
 *
 *		Destruct me
 */

XGViewFactory::~XGViewFactory()
{
	XGCritical lock(&GMutex);				// Lock entire function 
	
	XGViewFactory *foo;
	
	if (gList == this) gList = fNext;
	else {
		for (foo = gList; foo != NULL; foo = foo->fNext) {
			if (foo->fNext == this) {
				foo->fNext = fNext;
				break;
			}
		}
	}
}

/************************************************************************/
/*																		*/
/*	View Creation														*/
/*																		*/
/************************************************************************/

/*	XGViewFactory::_Create
 *
 *		Internal creation routine; this creates the specified view
 *	from the memory template specified. This creates the view and the
 *	children in that view
 */

XGView *XGViewFactory::_Create(long viewID, XGView *parent, XGSViewResHeader *tmpl)
{
	XGViewFactory *f;
	XGView *v;
	char buffer[6];
	char *tmp;
	XGArgStream stream((void *)(&(tmpl->fViewType)));
	unsigned char *data;
	XGSViewResHeader *vres;
	
	/*
	 *	Try to create this view
	 */
	
	for (f = gList; f != NULL; f = f->fNext) {
		v = f->CreateView(viewID,parent,stream);
		if (v) break;
	}
	
	/*
	 *	Post error message
	 */
	
	if (f == NULL) {
		tmp = (char *)&viewID;
#if OPT_MACOS == 1
		buffer[0] = tmp[0];
		buffer[1] = tmp[1];
		buffer[2] = tmp[2];
		buffer[3] = tmp[3];
#endif

#if OPT_WINOS == 1
		buffer[0] = tmp[3];
		buffer[1] = tmp[2];
		buffer[2] = tmp[1];
		buffer[3] = tmp[0];
#endif

		buffer[4] = 0;
		XPostError err("Unable to create view ('^S')",buffer);
		err.DisplayError();
		
		for (f = gList; f != NULL; f = f->fNext) {
			v = f->CreateView('view',parent,stream);
			if (v) break;
		}
		if (f == NULL) {
			throw XPostError("Unable to create 'view' class");
		}
	}
	
	/*
	 *	Create children
	 */
	
	if (tmpl->offchild) {
		data = ((unsigned char *)tmpl) + tmpl->offchild;
		for (;;) {
			vres = (XGSViewResHeader *)data;
			data += vres->offsibling;
			
			XGDispatch::_SetCTorParentDispatch(v);	/* Make sure created child */
			_Create(vres->fViewType,v,vres);		/* has correct dispatch parent */

			if (vres->offsibling == 0) break;		/* No more siblings */
		}
	}
	
	/*
	 *	Now send the 'size' message to force the children views to be
	 *	resized to their proper location
	 */
	
	v->DoSizeView();

	return v;
}

/*	XGViewFactory::_Create
 *
 *		This is the resource version of the creation engine
 */

XGView *XGViewFactory::_Create(XGView *parent, short resID)
{
	XGView *v;
	XGSViewResHeader *vh;

#if OPT_MACOS == 1
	Handle h;
	
	h = GetResource('VRes',resID);
	if (h == NULL) throw XPostError("Resource 'VRes'#^D missing",(long)resID);
	
	/*
	 *	Create this view
	 */
	
	HLock(h);
	vh = (XGSViewResHeader *)(*h);
	v = _Create(vh->fViewType,parent,vh);
	HUnlock(h);
	
	ReleaseResource(h);
#endif

#if OPT_WINOS == 1
	HRSRC find;
	HGLOBAL h;
	void *data;
	
	find = FindResource(_GInstance,MAKEINTRESOURCE(resID),"VRes");
	if (find == NULL) throw XPostError("Resource 'VRes'#^D missing",(long)resID);
	h = LoadResource(_GInstance,find);
	if (h == NULL) throw XPostError("Resource 'VRes'#^D missing",(long)resID);
	data = LockResource(h);
	
	/*
	 *	Create this view
	 */
	
	vh = (XGSViewResHeader *)data;
	v = _Create(vh->fViewType,parent,vh);
	
	/*
	 *	Release the resource
	 */
	
	UnlockResource(h);
	FreeResource(h);
#endif

	return v;
}
