/*	XGDialog.cpp
 *
 *		This is the dialog handling routines. This knows how to
 *	handle both modal and modeless dialogs.
 *
 *		A modeless dialog is really not much more than a window.
 *	A modal dialog is the same, except that the rest of the application
 *	is deactivated until the window is dismissed.
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XDialog.h>
#include <XApplication.h>

/************************************************************************/
/*																		*/
/*	Modal Dialog Declarations											*/
/*																		*/
/************************************************************************/

/*	XGDialog::XGDialog
 *
 *		Construct me
 */

XGDialog::XGDialog(XGSWindowInitRecord &x) : XGWindow(x)
{
	fDataPtr = NULL;
	fDialogProc = NULL;
	fTermValue = 0;
	fTerminate = false;
	fModal = false;
}

XGDialog::XGDialog(XGArgStream &x) : XGWindow(x)
{
	fDataPtr = NULL;
	fDialogProc = NULL;
	fTermValue = 0;
	fTerminate = false;
	fModal = false;
}

/*	XGDialog::~XGDialog
 *
 *		Destroy me
 */

XGDialog::~XGDialog()
{
	if (fDialogProc) {
		SendMessage(KEventDialogMessage,KEventDMEndDialog,NULL);
	}
	
	fDialogProc = NULL;
	fDataPtr = NULL;
	fTerminate = true;
}

/*	XGDialog::CloseWindow
 *
 *		An alternate way to close this; on the Microsoft Windows
 *	platform this is called when the window's close box is clicked
 */

bool XGDialog::CloseWindow()
{
	EndDialog(0);
	return true;
}

/************************************************************************/
/*																		*/
/*	Modeless creation/destruction										*/
/*																		*/
/************************************************************************/

/*	XGDialog::Create
 *
 *		Create this thing
 */

XGDialog *XGDialog::Create(short resID, XGPDialogProc proc, void *data)
{
	XGDialog *d;
	
	/*
	 *	Create my window
	 */
	
	d = (XGDialog *)XGViewFactory::_Create(NULL,resID);
	_SetCTorParentDispatch(NULL);
	
	/*
	 *	Now tell me what's up
	 */
	
	if (d) {
		d->fDataPtr = data;
		d->fDialogProc = proc;
		
		d->SendMessage(KEventDialogMessage,KEventDMInitDialog,NULL);
		if (d->fTerminate) {
			delete d;
			return NULL;
		}
	}
	return d;
}

/*	XGDialog::ModalDialog
 *
 *		This is the modal dialog version. On the Macintosh and on
 *	Windows I do all sorts of wierd magic stuff.
 */

long XGDialog::ModalDialog(short resID, XGPDialogProc proc, void *data)
{
	long l;
	XGDialog *d;
	
	/*
	 *	Create the thing
	 */
	
	d = Create(resID,proc,data);
	if (d == NULL) return 0;
	
	/*
	 *	Run the event loop until I'm told to terminate
	 */
	
#if OPT_WINOS == 1
	EnableWindow(_GMainWindow,FALSE);		// Disable the main window
#endif

	d->fModal = true;						// I'm a modal dialog
	for (;;) {
		_GApp->_RunOnce(d);
		if (d->fTerminate) break;
	}

#if OPT_WINOS == 1
	EnableWindow(_GMainWindow,TRUE);		// Re-enable it again
#endif
	
	/*
	 *	Now it's time to quit. Quit.
	 */
	
	l = d->fTermValue;
	delete d;
	return l;
}

/*	XGDialog::EndDialog
 *
 *		End this dialog. Destroy if we're a modeless
 */

void XGDialog::EndDialog(long v)
{
	fTerminate = true;
	fTermValue = v;
	
	if (!fModal) delete this;
}

/************************************************************************/
/*																		*/
/*	Dispatch Support													*/
/*																		*/
/************************************************************************/

/*	XGDialog::ReceiveDispatch
 *
 *		Process dispatch message
 */

long XGDialog::ReceiveDispatch(long msg, long arg, void *parg)
{
	long l;
	
	l = SendMessage(msg,arg,parg);
	if ((l == 0) && (!fTerminate)) {
		l = XGWindow::ReceiveDispatch(msg,arg,parg);
	}
	return l;
}

/************************************************************************/
/*																		*/
/*	Dispatch Support													*/
/*																		*/
/************************************************************************/

/*	XGDialog::SendMessage
 *
 *		Send a message to me
 */

long XGDialog::SendMessage(long msg, long arg, void *parg)
{
	if (fDialogProc) {
		return fDialogProc(this,fDataPtr,msg,arg,parg);
	} else {
		return 0;
	}
}
