/*	XGWindow.cpp
 *
 *		This is the basic window core class object
 */

/*  YAAF - Yet another application framework
 *  Copyright (C) 1997 William Edward Woody and In Phase Consulting
 *  
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or any
 *  later version.
 *  
 *  This library is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABIILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU Library General Public License for
 *  more details.
 *  
 *  You should have received a copy of the GNU Library General
 *  Public License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *  
 *  To contact the author, either e-mail me at
 *  woody@alumni.caltech.edu, or write to us at
 *  
 *          William Edward Woody
 *          In Phase Consulting
 *          1545 Ard Eevin Avenue
 *          Glendale, CA 91202
 */

#include <XConfig.h>
#include <XEvent.h>
#include <XError.h>
#include <XWindow.h>
#include <XMenu.h>
#include <XDataUtil.h>
#include <XCursorUtil.h>
#include <XApplication.h>
#include <XDocument.h>

#include <string.h>

#if (OPT_MACOS == 1) && (OPT_FLOATWINODW == 1)
#include "MacFloater.h"
#endif

#if (OPT_MACOS == 1) && (OPT_FLOATWINDOW == 0)
#define KNormalWindow		10			/* windowKind of *my* windows	*/
#endif

/************************************************************************/
/*																		*/
/*	Internal globals													*/
/*																		*/
/************************************************************************/

static short		GDefaultPosition;

/************************************************************************/
/*																		*/
/*	Construction/Destruction											*/
/*																		*/
/************************************************************************/

/*	XGWindow::XGWindow
 *
 *		This actually creates my window's internal data structures
 *	only. This simply initializes me--the actual creation process
 *	happens elsewhere.
 */

XGWindow::XGWindow(XGSWindowInitRecord &init) : XGFocus(this), XGView(init.fViewType)
{	
	/*
	 *	Initialize this window
	 */
	
	IInitRecord();
	IInitWindow(init);
	
	/*
	 *	The window is now invisible. Kludge create message by sending
	 *	it to my parent's receive method
	 */
	
	XGDispatch::ReceiveDispatch(KEventWindowCreate,init.windowID,(void *)this);

	/*
	 *	Show the window if it's hidden
	 */
	
	if (init.fVisible) {
		ShowWindow();
		SelectWindow();
	}
}

XGWindow::XGWindow(XGArgStream &stream) : XGFocus(this), XGView(stream)
{
	XGSWindowInitRecord w;
	
	w.fViewType = fViewType;
	
	w.windowType = stream.GetInteger();
	w.windowID = stream.GetInteger();
	w.fVisible = stream.GetBoolean();
	
	w.screenLeft = stream.GetInteger();
	w.screenTop = stream.GetInteger();
	w.screenRight = stream.GetInteger();
	w.screenBottom = stream.GetInteger();
	
	w.minx = stream.GetInteger();
	w.miny = stream.GetInteger();
	w.maxx = stream.GetInteger();
	w.maxy = stream.GetInteger();
	w.initx = stream.GetInteger();
	w.inity = stream.GetInteger();
	w.zoomx = stream.GetInteger();
	w.zoomy = stream.GetInteger();

	stream.GetString(sizeof(w.windowName),w.windowName);
	
	/*
	 *	Do the initialization sequence
	 */
	
	IInitRecord();
	IInitWindow(w);
	
	/*
	 *	The window is now invisible. Kludge create message by sending
	 *	it to my parent's receive method
	 */
	
	XGDispatch::ReceiveDispatch(KEventWindowCreate,w.windowID,(void *)this);

	/*
	 *	Show the window if it's hidden
	 */
	
	if (w.fVisible) {
		ShowWindow();
		SelectWindow();
	}
}


/*	XGWindow::~XGWindow
 *
 *		Unconditionally destroy this window
 */

XGWindow::~XGWindow()
{
	XGView *v;
	
	/*
	 *  Tell my document I'm gone
	 */
	
	XGDispatch::ReceiveDispatch(KEventWindowDestroy,fWinID,(void *)this);
	
	/*
	 *	Wipe out all of the children views -- this fixes a few
	 *	problems with views who insist on drawing on close, and
	 *	problems with the window's closure deleting view assets
	 */
	
	while (NULL != (v = GetChild())) delete v;
	
	/*
	 *  Close this window itself
	 */
	
#if OPT_MACOS == 1
	if (fWindow.tw != NULL) {
#if OPT_FLOATWINDOW == 1
		::MCloseWindow((WindowPtr)&fWindow.w);
#else
		::CloseWindow((WindowPtr)&fWindow.w);
#endif
	}
#endif

#if OPT_WINOS == 1
	if (fViewHWND) {
		SetProp(fViewHWND,_GWindowClass,NULL);
		if (_GMDIWindow) {
			::SendMessage(_GMDIWindow,WM_MDIDESTROY,(WPARAM)fViewHWND,NULL);
		} else {
			DestroyWindow(fViewHWND);
		}
		fViewHWND = NULL;							// and forget the window
	}
#endif
}

/************************************************************************/
/*																		*/
/*	Create/Destroy														*/
/*																		*/
/************************************************************************/

/*	XGWindow::Create
 *
 *		Create window from resource. This is a bit wierd, but this
 *	wrappers the XGViewFactory creation routine.
 */

XGWindow *XGWindow::Create(short resID)
{
	XGWindow *tw;
	
	/*
	 *	Create my window
	 */
	
	tw = (XGWindow *)XGViewFactory::_Create(NULL,resID);
	_SetCTorParentDispatch(NULL);
	return tw;
}

/*	XGWindow::Create
 *
 *		The locked resource version of the above
 */

XGWindow *XGWindow::Create(XGSWindowInitRecord &w)
{
	XGWindow *tw;
	
	tw = new XGWindow(w);
	_SetCTorParentDispatch(NULL);
	return tw;
}

/*	XGWindow::CloseWindow
 *
 *		Handle closing this window. This queries the contained document
 *	and processes the close after saving the file
 */

bool XGWindow::CloseWindow()
{
	char name[256];
	int result;
#if OPT_WINOS == 1
	char buffer[256];
#endif

	/*
	 *	Send the predestroy message. The document
	 */
	
	XGDispatch::ReceiveDispatch(KEventWindowPreDestroy,fWinID,(void *)this);
	
	/*
	 *	If this is dirty, put up the save message
	 */
	
	if (IsDirty()) {
		/*
		 *	This window is dirty. Determine if the user wants to save
		 *	the window.
		 */
		
		GetWindowName(name);
		NormalCursor();

#if OPT_MACOS == 1
		ParamText(c2pstr(name),NULL,NULL,NULL);
		result = Alert(256,NULL);					//###
		if (result == 1) {
			if (!DoSave()) return 0;				// Failed to save
		} else if (result == 2) return 0;			// User canceled.
#endif

#if OPT_WINOS == 1
		wsprintf(buffer,"Should the contents of %s be saved before the window is closed?",name);
		result = MessageBox(fViewHWND,buffer,"Save",MB_YESNOCANCEL | MB_ICONQUESTION);
		if (result == IDYES) {
			if (!DoSave()) return 0;				// Failed to save
		} else if (result == IDCANCEL) return 0;	// User canceled
#endif
	}
	
	/*
	 *	We got here. Kill the window and return success.
	 */
	 
	delete this;
	return 1;
}

/************************************************************************/
/*																		*/
/*	Dirty/Save Support													*/
/*																		*/
/************************************************************************/

/*	XGWindow::IsDirty
 *
 *		Is this dirty? This by default passes up to the document
 */

bool XGWindow::IsDirty()
{
	if (fDocument) return fDocument->IsDirty();
	else return false;
}

/*	XGWindow::DoSave
 *
 *		Do save. This by default passes up to the document
 */

bool XGWindow::DoSave()
{
	if (fDocument) return fDocument->DoSave();
	else return true;
}

/************************************************************************/
/*																		*/
/*	Window Min/Max stuff												*/
/*																		*/
/************************************************************************/

/*	XGWindow::MinWindowSize
 *
 *		Default: return what I read in
 */

Point XGWindow::MinWindowSize()
{
	return fMin;
}

/*	XGWindow::MaxWindowSize
 *
 *		Default: return what I read in
 */

Point XGWindow::MaxWindowSize()
{
	return fMax;
}

/*	XGWindow::InitWindowSize
 *
 *		Default: return what I read in
 */

Point XGWindow::InitWindowSize()
{
	return fInit;
}

/*	XGWindow::ZoomWindowSize
 *
 *		Default: return what I read in
 */

Point XGWindow::ZoomWindowSize()
{
	return fZoom;
}

/*	XGWindow::_LimitWindowSize
 *
 *		Limit the window size
 */

void XGWindow::_LimitWindowSize(Point *pt)
{
	Point n;
	
	n = MinWindowSize();
	if (pt->h != -1) {
		if (n.h != -1) {
			if (pt->h < n.h) pt->h = n.h;
		} else {
			if (pt->h < 75) pt->h = 75;
		}
	}
	if (pt->v != -1) {
		if (n.v != -1) {
			if (pt->v < n.v) pt->v = n.v;
		} else {
			if (pt->v < 125) pt->v = 125;
		}
	}
	
	n = MaxWindowSize();
	if ((pt->h != -1) && (n.h != -1)) if (pt->h > n.h) pt->h = n.h;
	if ((pt->v != -1) && (n.v != -1)) if (pt->v > n.v) pt->v = n.v;
}

/************************************************************************/
/*																		*/
/*	View Overrides														*/
/*																		*/
/************************************************************************/

/*	XGWindow::DoDrawView
 *
 *		This overrides the default view behaviour to be obnoxious--this
 *	forces the view to do nothing.
 */

void XGWindow::DoDrawView(Rect)
{
}

/************************************************************************/
/*																		*/
/*	Window Attributes													*/
/*																		*/
/************************************************************************/

/*	XGWindow::GetWindowName
 *
 *		Get the name of this window
 */

void XGWindow::GetWindowName(char *name)
{
#if OPT_MACOS == 1
	GetWTitle((WindowPtr)&fWindow.w,(unsigned char *)name);
	p2cstr((unsigned char *)name);
#endif

#if OPT_WINOS == 1
	GetWindowText(fViewHWND,name,256);
#endif
}

/*	XGWindow::SetWindowName
 *
 *		Set the window name
 */

void XGWindow::SetWindowName(char *name)
{
#if OPT_MACOS == 1
	unsigned char buffer[256];

	SetWTitle((WindowPtr)&fWindow.w,cvtc2p(buffer,name));
#endif

#if OPT_WINOS == 1
	SetWindowText(fViewHWND,name);
#endif

	XGDispatch::ReceiveDispatch(KEventWindowName,fWinID,(void *)this);
}

/*	XGWindow::GetMenuBarID
 *
 *		Get the resource ID for this window
 */

short XGWindow::GetMenuBarID(void)
{
	return KDefaultMenu;
}

/************************************************************************/
/*																		*/
/*	Window Attributes													*/
/*																		*/
/************************************************************************/

/*	XGWindow::ShowWindow
 *
 *		Show this window.
 */

void XGWindow::ShowWindow()
{
	if (!IsShowWindow()) {
	
#if OPT_MACOS == 1
#if OPT_FLOATWINDOW == 1
		::MShowWindow((WindowPtr)&fWindow.w);
#else /* OPT_FLOATWINDOW != 1 */
		::ShowWindow((WindowPtr)&fWindow.w);
#endif /* OPT_FLOATWINODW */
#endif /* OPT_MACOS */

#if OPT_WINOS == 1
		::ShowWindow(fViewHWND,SW_SHOW);
#endif

	}
}

/*	XGWindow::HideWindow
 *
 *		Show this window.
 */

void XGWindow::HideWindow()
{
	if (IsShowWindow()) {
	
#if OPT_MACOS == 1
#if OPT_FLOATWINDOW == 1
		::MHideWindow((WindowPtr)&fWindow.w);
#else  /* OPT_FLOATWINDOW != 1 */
		::HideWindow((WindowPtr)&fWindow.w);
#endif /* OPT_FLOATWINDOW */
#endif /* OPT_MACOS */

#if OPT_WINOS == 1
		::ShowWindow(fViewHWND,SW_HIDE);
#endif

	}
}

/*	XGWindow::IsShowWindow
 *
 *		Is this window visible?
 */

bool XGWindow::IsShowWindow()
{
#if OPT_MACOS == 1
	return fWindow.w.visible;
#endif

#if OPT_WINOS == 1
	return IsWindowVisible(fViewHWND);
#endif
}

/*	XGWindow::GetFrontWindow
 *
 *		Get the front window
 */

XGWindow *XGWindow::GetFrontWindow()
{
#if OPT_MACOS == 1
	WindowPtr w;
	
#if OPT_FLOATWINDOW == 1
	w = MFrontWindow();
#else
	w = FrontWindow();
#endif

	return _GetXGWindow(w);
#endif

#if OPT_WINOS == 1
	HWND w;
	
	/* If this is an SDI app, we have only one window */
	if (_GMDIWindow == NULL) {
		return XGWindow::_GetXGWindow(_GMainWindow);
	}
	
	/* It's an MDI window. Find the frontmost child to the MDI pane */
	w = ::GetWindow(_GMDIWindow,GW_CHILD);
	while (w != NULL) {
		if (::GetProp(w,_GWindowClass)) {
			if (::IsWindowVisible(w)) break;
		}
		w = ::GetWindow(w,GW_HWNDNEXT);
	}
	if (w == NULL) return NULL;
	return XGWindow::_GetXGWindow(w);
#endif
}

/*	XGWindow::SelectWindow
 *
 *		Bring this window frontmost
 */

void XGWindow::SelectWindow()
{
#if OPT_MACOS == 1
#if OPT_FLOATWINDOW == 1
	MSelectWindow((WindowPtr)&fWindow.w);
#else /* OPT_FLOATWINDOW != 1 */
	::SelectWindow((WindowPtr)&fWindow.w);
#endif
#endif

#if OPT_WINOS == 1
	/* WM_MDIACTIVATE arguments same for Win16 and Win32 in this mode */
	if (_GMDIWindow) {
		::SendMessage(_GMDIWindow,WM_MDIACTIVATE,(WPARAM)fViewHWND,NULL);
		if (IsIconic()) ::SendMessage(_GMDIWindow,WM_MDIRESTORE,(WPARAM)fViewHWND,NULL);
	} else {
		::ShowWindow(fViewHWND,SW_RESTORE);
	}
#endif
}

/*	XGWindow::GetScreenSize
 *
 *		Return the size of the main screen (minus the menu bar)
 */

void XGWindow::GetScreenSize(Rect *r)
{
#if OPT_MACOS == 1
	WindowPtr w;
		
	GetWMgrPort(&w);
	*r = w->portRect;
	r->top += LMGetMBarHeight();
#endif

#if OPT_WINOS == 1
	if (_GMDIWindow) {
		::GetClientRect(_GMDIWindow,r);
	} else {
		r->left   = 0;
		r->top    = 0;
		r->right  = ::GetSystemMetrics(SM_CXFULLSCREEN);
		r->bottom = ::GetSystemMetrics(SM_CYFULLSCREEN);
	}
#endif
}

/************************************************************************/
/*																		*/
/*	Window Size, Content Size Manipulation								*/
/*																		*/
/************************************************************************/

/*	XGWindow::GetContentSize
 *
 *		Return the size of the content area
 */

Point XGWindow::GetContentSize(void) const
{
#if OPT_MACOS == 1
	Point pt;
	
	pt.h = fWindow.w.port.portRect.right - fWindow.w.port.portRect.left;
	pt.v = fWindow.w.port.portRect.bottom - fWindow.w.port.portRect.top;
	
	return pt;
#endif

#if OPT_WINOS == 1
	Point pt;
	Rect r;
	
	GetClientRect(fViewHWND,&r);
	
	pt.h = r.right - r.left;
	pt.v = r.bottom - r.top;
	
	return pt;
#endif
}

/*	XGWindow::GetContentRect
 *
 *		Get the content rectangle
 */

Rect XGWindow::GetContentRect() const
{
	Rect r;
	Point size;
	
	size = GetContentSize();
	r.left = 0;
	r.top = 0;
	r.right = size.h;
	r.bottom = size.v;
	
	return r;
}

/*	XGWindow::SetWindowSize
 *
 *		Resize this window
 */

void XGWindow::SetWindowSize(Point size)
{
	_LimitWindowSize(&size);
	
#if OPT_MACOS == 1
	size.h -= 2;
	size.v -= 22;					// Convert to the internal size
	SizeWindow((WindowPtr)&fWindow.w,size.h,size.v,1);

	_ProcessSizeEvent();
#endif

#if OPT_WINOS == 1
	::SetWindowPos(fViewHWND,NULL,0,0,size.h,size.v,SWP_NOMOVE | SWP_NOZORDER);
#endif
}

/*	XGWindow::SetWindowPos
 *
 *		Set the window's location
 */

void XGWindow::SetWindowPos(Point pos)
{
#if OPT_MACOS == 1
	pos.h += 1;
	pos.v += 21;
	MoveWindow((WindowPtr)&fWindow.w,pos.h,pos.v,0);
#endif

#if OPT_WINOS == 1
	::SetWindowPos(fViewHWND,NULL,pos.h,pos.v,0,0,SWP_NOSIZE | SWP_NOZORDER);
#endif
}

/*	XGWindow::SetWindowLocation
 *
 *		Set the window's global location
 */

void XGWindow::SetWindowLocation(Rect r)
{
	Point pt;
	
	pt.h = r.right - r.left;
	pt.v = r.bottom - r.top;
	_LimitWindowSize(&pt);
	r.right = r.left + pt.h;
	r.bottom = r.top + pt.v;
	
#if OPT_MACOS == 1
	r.top += 21;									// Convert to content size
	r.left += 1;
	r.right -= 1;
	r.bottom -= 1;
#endif
	
#if OPT_MACOS == 1
	ShowHide((WindowPtr)&fWindow.w,0);				// Temporary hide
	SizeWindow((WindowPtr)&fWindow.w,r.right-r.left,r.bottom-r.top,0);
	MoveWindow((WindowPtr)&fWindow.w,r.left,r.top,0);
	ShowHide((WindowPtr)&fWindow.w,1);				// show it again

	_ProcessSizeEvent();
#endif

#if OPT_WINOS == 1
	::SetWindowPos(fViewHWND,NULL,r.left,r.top,r.right-r.left,r.bottom-r.top,SWP_NOZORDER);
#endif
}

/*	XGWindow::GetWindowSize
 *
 *		Get the window's current size
 */

Point XGWindow::GetWindowSize() const
{
	Rect r;
	Point pt;
	
#if OPT_MACOS == 1
	r = fWindow.w.port.portRect;
	pt.h = r.right - r.left + 2;
	pt.v = r.bottom - r.top + 22;
#endif

#if OPT_WINOS == 1
	GetWindowRect(fViewHWND,&r);
	pt.h = r.right - r.left;
	pt.v = r.bottom - r.top;
#endif

	return pt;
}

/*	XGWindow::GetWindowPos
 *
 *		Get the window's current location
 */

Point XGWindow::GetWindowPos() const
{
	Point pt;
	
#if OPT_MACOS == 1
	const Rect *r;
	CGrafPtr port;
#endif

#if OPT_WINOS == 1
	Rect r;
#endif

#if OPT_MACOS == 1
	if (fWindow.w.port.portBits.rowBytes & 0x8000) {	// Color version
		port = (CGrafPtr)&fWindow.w.port;
		r = &(**(port->portPixMap)).bounds;
	} else {
		r = &fWindow.w.port.portBits.bounds;
	}
	pt.h = fWindow.w.port.portRect.left - r->left - 1;
	pt.v = fWindow.w.port.portRect.top- r->top - 21;
#endif

#if OPT_WINOS == 1
	GetWindowRect(fViewHWND,&r);
	pt.h = r.left;
	pt.v = r.top;
#endif
	
	return pt;
}


/************************************************************************/
/*																		*/
/*	Internal Miscellaneous Stuff										*/
/*																		*/
/************************************************************************/

/*	XGWindow::_IsActiveWindow
 *
 *		Is this window active?
 */

bool XGWindow::_IsActiveWindow() const
{
#if OPT_MACOS == 1
	return fWindow.w.hilited;
#endif

#if OPT_WINOS == 1
	return (_GetHWND() == GetFrontWindow());
#endif
}

/*	XGWindow::_CapturedView
 *
 *		What is captured?
 */

XGView *XGWindow::_CapturedView() const
{
#if OPT_MACOS == 1
	return fCapture;
#endif

#if OPT_WINOS == 1
	return XGView::_GetXGView(GetCapture());
#endif
}


/*	XGWindow::_ContentToSize
 *
 *		Convert from the content area to the outer size
 */

void XGWindow::_ContentToSize(Rect *r) const
{
	if (fWinType == KWinTypeDocument) {
		r->left--;
		r->right++;
		r->bottom++;
		r->top -= 21;
	}
}

void XGWindow::_ContentToSize(Point *pt) const
{
	if (fWinType == KWinTypeDocument) {
		if (pt->h != -1) pt->h += 2;
		if (pt->v != -1) pt->v += 22;
	}
}

/*	XGWindow::_SizeToContent
 *
 *		Do the inverse
 */

void XGWindow::_SizeToContent(Rect *r) const
{
	if (fWinType == KWinTypeDocument) {
		r->left++;
		r->right--;
		r->bottom--;
		r->top += 21;
	}
}

void XGWindow::_SizeToContent(Point *pt) const
{
	if (fWinType == KWinTypeDocument) {
		if (pt->h != -1) pt->h -= 2;
		if (pt->v != -1) pt->v -= 22;
	}
}

/************************************************************************/
/*																		*/
/*	Macintosh stuff														*/
/*																		*/
/************************************************************************/

#if OPT_MACOS == 1

/*	XGWindow::_ResetDrawing
 *
 *		Handle resetting the drawing
 */
 
void XGWindow::_ResetDrawing()
{
	::SetPort((WindowPtr)&fWindow.w);
	SetOrigin(0,0);
	::ClipRect(&fWindow.w.port.portRect);
}

/*	XGWindow::_GetXGWindow
 *
 *		Get the window from this window pointer
 */

XGWindow *XGWindow::_GetXGWindow(WindowPtr w)
{
	if (w == NULL) return NULL;

#if OPT_FLOATWINDOW == 1
	if ((((WindowPeek)w)->windowKind != KNormalWindow) &&
		(((WindowPeek)w)->windowKind != KFloatingWindow)) return NULL;
#else
	if (((WindowPeek)w)->windowKind != KNormalWindow) return NULL;
#endif

	return (((_MyWindowRecord *)w)->tw);
}

/*	XGWindow::_ProcessUpdateEvent
 *
 *		Call the drawing routine for the view inside of this window
 */

void XGWindow::_ProcessUpdateEvent(Rect r)
{
	_DrawViewHierarchy(r);
}

/*	XGWindow::_ProcessActivateEvent
 *
 *		Handle the activation/deactivation of this window
 */

void XGWindow::_ProcessActivateEvent(bool flag)
{
	_ActivateViewHierarchy(flag);
}

/*	XGWindow::_ProcessMouseDown
 *
 *		Handle mouse down event. Set the capture bit if the mouse becomes
 *	captured in any interesting way.
 */

bool XGWindow::_ProcessMouseDown(Point loc, short flags)
{
	XGView *view;
	
	if (NULL == (view = FindViewByPoint(loc))) return 0;
	
	view->GlobalToView(&loc);
	if (view->DoMouseDown(loc,flags)) {
		fCapture = view;
		return 1;
	} else return 0;
}

/*	XGWindow::_ProcessMouseMove
 *
 *		Handle mouse moved event
 */

void XGWindow::_ProcessMouseMove(Point pt, short flags)
{
//	extern XGWindow *_GTopWindow;
	XGView *view;	
	short flag;
	XGWindow *tw;
	WindowPtr w;
	
	flag = 0;
	if (fCapture) {
		fCapture->GlobalToView(&pt);
		fCapture->DoMouseMove(pt,flags | KCapturedMouse);
	} else {
		if (NULL == (view = FindViewByPoint(pt))) return;
		tw = view->GetWindow();
		
		w = tw->_GetGrafPtr();
		if (!((WindowPeek)w)->hilited) {	/* Not over anything */
			NormalCursor();
			return;
		}

		view->GlobalToView(&pt);
		view->DoMouseMove(pt,flags);
	}
}

/*	XGWindow::_ProcessMouseUp
 *
 *		Handle mouse up event
 */

void XGWindow::_ProcessMouseUp(Point pt, short flags)
{
	XGView *view;
	
	if (fCapture) {
		fCapture->GlobalToView(&pt);
		fCapture->DoMouseUp(pt,flags | KCapturedMouse);
		fCapture = NULL;
	} else {
		if (NULL == (view = FindViewByPoint(pt))) return;
		view->GlobalToView(&pt);
		view->DoMouseUp(pt,flags);
	}
}

#endif

/*	XGWindow::_ProcessSizeEvent
 *
 *		Handle resizing the window by forcing the view to resize
 */

void XGWindow::_ProcessSizeEvent(void)
{
	SetLocation(GetContentRect());				// (! size is always from WindowRecord)
}

/************************************************************************/
/*																		*/
/*	Low level stuff														*/
/*																		*/
/************************************************************************/

#if OPT_WINOS == 1

/*	XGWindow::_GetXGWindow
 *
 *		Get the XGWindow object given the HWND
 */

XGWindow *XGWindow::_GetXGWindow(HWND w)
{
	if (w == NULL) return NULL;
	else return (XGWindow *)GetProp(w,_GWindowClass);
}

/*	XGWindow::IsIconic
 *
 *		Is this thing iconic?
 */

bool XGWindow::IsIconic()
{
	return ::IsIconic(fViewHWND);
}

#endif

/************************************************************************/
/*																		*/
/*	How To Initialize A Window											*/
/*																		*/
/************************************************************************/

/*	XGWindow::IInitRecord
 *
 *		Initialize my data structure
 */

void XGWindow::IInitRecord()
{
	fDocument = NULL;

#if OPT_MACOS == 1
	fCapture = NULL;
	fWindow.tw = NULL;
#endif

#if OPT_WINOS == 1
	fViewHWND = NULL;
#endif

	fMin.h = 0;
	fMin.v = 0;
	fMax.h = 0;
	fMax.v = 0;
	fInit.h = 0;
	fInit.v = 0;
	fZoom.h = 0;
	fZoom.v = 0;
	
	fWinID = 0;
}

/*	XGWindow::IInitWindow
 *
 *		Initialize this window from a WinResRecord object
 */

void XGWindow::IInitWindow(XGSWindowInitRecord &w)
{
	Rect r;
	char buffer[256];
	
	/*
	 *	Copy data structures
	 */
	 
	fMin.h = w.minx;
	fMin.v = w.miny;
	fMax.h = w.maxx;
	fMax.v = w.maxy;
	fInit.h = w.initx;
	fInit.v = w.inity;
	fZoom.h = w.zoomx;
	fZoom.v = w.zoomy;
	
	fWinID = w.windowID;
	fWinType = w.windowType;	
	switch (fWinType) {
		case KWinTypeDocument:
			fAttributes = 0;
			break;
		case KWinTypeModal:
			fAttributes = KAttrNoClose | KAttrNoResize | KAttrNoMenu;
			break;
		case KWinTypeMovableModal:
			fAttributes = KAttrNoClose | KAttrNoResize | KAttrNoMenu;
			break;
#if OPT_FLOATWINDOW == 1
		case KWinTypeFloating:
			fAttributes = KAttrNoClose | KAttrNoResize | KAttrHideOnSwitch | KAttrFloater;
			break;
#endif
		default:
			fAttributes = 0;
			break;
	}
	
	/*
	 *	Figure out my name
	 */
	 
	strcpy(buffer,w.windowName);			// use default
	
	/*
	 *	Now initialize my window using the ICreateWindow object
	 */
	
	r.left = w.screenLeft;
	r.right = w.screenRight;
	r.top = w.screenTop;
	r.bottom = w.screenBottom;
	XGWindow::ICreateWindow(buffer,&r);
}

/*	XGWindow::ICreateWindow
 *
 *		This is the code to create the window based on the information
 *	just loaded
 */

void XGWindow::ICreateWindow(char *name, Rect *size)
{
#if OPT_MACOS == 1
	Rect s;
	Point pt;
	unsigned char buffer[256];
	short whichProc;
	long color;
#endif

#if OPT_WINOS == 1
	MDICREATESTRUCT mcs;
	Rect s;
	Point pt;
	extern XGWindow *_GWindow;
//	Rect sizerect;
#endif
	
	/*
	 *	Figure out what the default size is going to be with this thing
	 */
	
	if (EmptyRect(size)) size = NULL;			// Kludge for default size
	if (size == NULL) {
#if OPT_MACOS == 1
		s.top = GDefaultPosition * 15 + LMGetMBarHeight();
		s.left = GDefaultPosition * 15 + 5;
#endif

#if OPT_WINOS == 1
		s.top = GDefaultPosition * (GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION));
		s.left = GDefaultPosition * (GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION));
#endif
		GDefaultPosition++;
		if (GDefaultPosition >= 5) GDefaultPosition = 0;
		pt = InitWindowSize();
		if (pt.h == -1) pt.h = 320;
		if (pt.v == -1) pt.v = 240;
		s.right = s.left + pt.h;
		s.bottom = s.top + pt.v;
	} else s = *size;
	
	// kludge for offset even though specified size
	if ((s.left == 0) && (s.top == 0)) {
#if OPT_MACOS == 1
		s.top = GDefaultPosition * 15 + LMGetMBarHeight();
		s.left = GDefaultPosition * 15 + 5;
#endif

#if OPT_WINOS == 1
		s.top = GDefaultPosition * (GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION));
		s.left = GDefaultPosition * (GetSystemMetrics(SM_CYFRAME) + GetSystemMetrics(SM_CYCAPTION));
#endif
		GDefaultPosition++;
		if (GDefaultPosition >= 5) GDefaultPosition = 0;
		s.bottom += s.top;
		s.right += s.left;
	}

	// now limit the size to the size routine provided
	pt.h = s.right - s.left;
	pt.v = s.bottom - s.top;
	_LimitWindowSize(&pt);
	s.right = s.left + pt.h;
	s.bottom = s.top + pt.v;

#if OPT_MACOS == 1
	/*
	 *	Open the window itself [Macintosh version]
	 */
	 
	_SizeToContent(&s);					// Convert to internal content size
	
	Gestalt(gestaltQuickdrawVersion,&color);
	if (color >= 0x100) color = 1;		// Do I have color?
	else color = 0;
	
	fWindow.tw = this;					// Get the backpointer correct.	

	c2pstr(strcpy((char *)buffer,name));
	
	if (fWinType == KWinTypeDocument) whichProc = 8;			// zoom proc
	else if (fWinType == KWinTypeModal) whichProc = 1;			// modal
	else if (fWinType == KWinTypeMovableModal) whichProc = 5;	// movemodal
#if OPT_FLOATWINDOW == 1
	else if (fWinType == KWinTypeFloating) whichProc = 2048;	// floater
#endif
	else whichProc = noGrowDocProc;

#if OPT_FLOATWINDOW == 1
	if (color) {
		if (NULL == ::MNewCWindow((Ptr)&fWindow.w,&s,buffer,0,whichProc,(WindowPtr)-1,!(_Attributes() & KAttrNoClose),0L)) {
			fWindow.tw = NULL;			// Clear the backpointer
			throw XPostError("Unable to open window: out of memory");
		}
	} else {
		if (NULL == ::MNewWindow((Ptr)&fWindow.w,&s,buffer,0,whichProc,(WindowPtr)-1,!(_Attributes() & KAttrNoClose),0L)){
			fWindow.tw = NULL;			// Clear the backpointer
			throw XPostError("Unable to open window: out of memory");
		}
	}
#else /* OPT_FLOATWINDOW == 0 */
	if (color) {
		if (NULL == ::NewCWindow((Ptr)&fWindow.w,&s,buffer,0,whichProc,(WindowPtr)-1,!(_Attributes() & KAttrNoClose),0L)) {
			fWindow.tw = NULL;			// Clear the backpointer
			throw XPostError("Unable to open window: out of memory");
		}
	} else {
		if (NULL == ::NewWindow((Ptr)&fWindow.w,&s,buffer,0,whichProc,(WindowPtr)-1,!(_Attributes() & KAttrNoClose),0L)){
			fWindow.tw = NULL;			// Clear the backpointer
			throw XPostError("Unable to open window: out of memory");
		}
	}
#endif
 
    /*
     *  Initialize drawing port
     */
	
	::SetPort((WindowPtr)&fWindow.w);
	::TextSize(9);						// Set a reasonable default font size
#if OPT_FLOATWINDOW == 1
	fWindow.w.windowKind = (_Attributes() & KAttrFloater) ? KFloatingWindow : KNormalWindow;
#else
	fWindow.w.windowKind = KNormalWindow;
#endif /* OPT_FLOATWINDOW */

#endif /* OPT_MACOS */

#if OPT_WINOS == 1

	if ((_GMDIWindow) && (GetViewType() == 'wind')) {
	    mcs.szClass = _GWindowClass;
	    mcs.szTitle = name;
	    mcs.hOwner = _GInstance;
		mcs.x = s.left;
		mcs.y = s.top;
		mcs.cx = s.right - s.left;
		mcs.cy = s.bottom - s.top;
	
	    mcs.style = 0;
	    mcs.lParam = NULL;						// Pass this on
		_GWindow = this;
		fViewHWND = (HWND)::SendMessage(_GMDIWindow,WM_MDICREATE,0,(LPARAM)(MDICREATESTRUCT FAR *)&mcs);
		_GWindow = NULL;
		if (fViewHWND == NULL) {
			throw XPostError("Unable to create window");
		}
	} else {
		_GWindow = this;
		
		if (GetViewType() == 'wind') {
			// Window
			fViewHWND = CreateWindow(_GWindowClass,
									name,
									WS_OVERLAPPEDWINDOW | // WS_CLIPCHILDREN | 
								   		WS_MAXIMIZEBOX | WS_MINIMIZEBOX,
									s.left,
									s.top,
									s.right-s.left,
									s.bottom-s.top,
									NULL,
									NULL,
									_GInstance,
									NULL);
		} else {
			// Dialog
			fViewHWND = CreateWindow(_GDlogClass,
									name,
									WS_CLIPCHILDREN | 
								   		WS_CAPTION | WS_POPUPWINDOW | WS_BORDER,
									s.left,
									s.top,
									s.right-s.left,
									s.bottom-s.top,
									_GMainWindow,
									NULL,
									_GInstance,
									NULL);
		}
		_GWindow = NULL;
		
		if (fViewHWND == NULL) {
			throw XPostError("Unable to create window");
		}
	}

//  }

#endif
	
	/*
	 *	Handle the initialization process of the subclass objects
	 */
	
	XGView::SetLocation(GetContentRect());	// initialize interior location
}

